#!/usr/bin/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the OverlayInput WorkflowModule."""

from __future__ import print_function
from __future__ import absolute_import
import os
import shutil
import tempfile
import unittest
from mock import patch, mock_open, call, MagicMock as Mock

from parameterized import parameterized

from DIRAC import gLogger, S_OK, S_ERROR
from ILCDIRAC.Workflow.Modules.OverlayInput import OverlayInput
from Tests.Utilities.GeneralUtils import assertEqualsImproved, \
    assertDiracFailsWith, assertDiracSucceeds, assertDiracSucceedsWith, \
    assertDiracSucceedsWith_equals
from Tests.Utilities.FileUtils import FileUtil

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Workflow.Modules.OverlayInput'
MODULEBASE_NAME = 'ILCDIRAC.Workflow.Modules.ModuleBase'

gLogger.setLevel("ERROR")
gLogger.showHeaders(True)


def cleanup(tempdir):
  """Remove files after run."""
  try:
    shutil.rmtree(tempdir)
  except OSError:
    pass


def createFile(*_args, **_kwargs):
  """create a file with filename if given."""
  with open("overlayFile.slcio", "w") as oFile:
    oFile.write("Somecontent")


@patch("%s.getProxyInfoAsString" % MODULEBASE_NAME, new=Mock(return_value=S_OK()))
@patch("DIRAC.Core.Security.ProxyInfo.getProxyInfoAsString", new=Mock(return_value=S_OK()))
@patch("%s.FileCatalogClient" % MODULE_NAME, new=Mock(return_value=S_OK()))
@patch("%s.Operations" % MODULE_NAME, new=Mock(return_value=S_OK()))
@patch("%s.DataManager" % MODULE_NAME, new=Mock(return_value=S_OK()))
class TestOverlayEos(unittest.TestCase):
  """test Getting Overlay files from CERN EOS.

  Make sure the copying command is properly formated and uses the correct path to the eos instance
  """

  def assertIn(self, *args, **kwargs):
    """make this existing to placate pylint."""
    return super(TestOverlayEos, self).assertIn(*args, **kwargs)

  def setUp(self):
    self.tmpdir = tempfile.mkdtemp("", dir="./")
    os.chdir(self.tmpdir)
    self.over = OverlayInput()
    self.over.applicationLog = "testOver.log"

  def tearDown(self):
    os.chdir("../")
    cleanup(self.tmpdir)

  @patch("%s.shellCall" % MODULE_NAME, new=Mock(side_effect=createFile))
  def test_overlayinput_getEosFile_lfn_success(self):
    """test success when getting an lfn to copy from eos."""
    testLFN = "/lfn/to/overlay/overlayFile.slcio"
    res = self.over.getEOSFile(testLFN)
    print(res)
    print("self result", self.over.result)
    assertDiracSucceedsWith_equals(res, os.path.basename(testLFN), self)
    with open("overlayinput.sh") as overscript:
      self.assertIn("xrdcp -s root://eospublic.cern.ch//eos/experiment/clicdp/grid%s" % testLFN, overscript.read())

  @patch("%s.shellCall" % MODULE_NAME, new=Mock(side_effect=createFile))
  def test_overlayinput_getEosFile_fullpath_success(self):
    """test that we don't predent if we get a fullpath for eos, however that might happen."""
    testLFN = "/eos/experiment/clicdp/grid/lfn/to/overlay/overlayFile.slcio"
    res = self.over.getEOSFile(testLFN)
    print(res)
    print("self result", self.over.result)
    assertDiracSucceedsWith_equals(res, os.path.basename(testLFN), self)
    with open("overlayinput.sh") as overscript:
      self.assertIn("xrdcp -s root://eospublic.cern.ch/%s" % testLFN, overscript.read())

  @patch("%s.shellCall" % MODULE_NAME, new=Mock())
  def test_overlayinput_getEosFile_lfn_failure(self):
    """test failure of copy command, that is no ouputfile present after copying."""
    testLFN = "/lfn/to/overlay/overlayFile.slcio"
    res = self.over.getEOSFile(testLFN)
    print(res)
    print("self result", self.over.result)
    assertDiracFailsWith(res, 'Failed', self)
    with open("overlayinput.sh") as overscript:
      self.assertIn("xrdcp -s root://eospublic.cern.ch//eos/experiment/clicdp/grid%s" % testLFN, overscript.read())

# pylint: disable=too-many-public-methods


class TestOverlayUnittests(unittest.TestCase):
  """Tests the Overlayinput class."""

  GOOD_EXIT = 'Input variables resolved'

  def setUp(self):
    self.over = OverlayInput()
    self.over.detectormodel = 'testdetectorv2000'
    self.over.energytouse = '200TeV'
    self.over.BXOverlay = 100
    self.over.NbSigEvtsPerJob = 20

  def test_applicationSpecificInputs(self):
    # Compare to this. Get outside of patch, else constructor fails
    reference = OverlayInput()
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=2)):
      result = self.over.applicationSpecificInputs()
      assertDiracSucceedsWith_equals(result, TestOverlayUnittests.GOOD_EXIT, self)
      # Assert nothing has been changed, except the values in setUp (and
      # DataManager/FileCatalogClient since theyre created anew for every
      # object)
      assertEqualsImproved((self.over.enable, self.over.STEP_NUMBER, self.over.applicationName,
                            self.over.curdir, self.over.applicationLog, self.over.printoutflag,
                            self.over.prodid, self.over.detector, self.over.energy, self.over.nbofeventsperfile,
                            self.over.lfns, self.over.nbfilestoget, self.over.BkgEvtType, self.over.ggtohadint,
                            self.over.nbsigeventsperfile, self.over.nbinputsigfile, self.over.site,
                            self.over.useEnergyForFileLookup, self.over.machine, self.over.pathToOverlayFiles,
                            ), (reference.enable, reference.STEP_NUMBER, reference.applicationName,
                                reference.curdir, reference.applicationLog, reference.printoutflag, reference.prodid,
                                reference.detector, reference.energy, reference.nbofeventsperfile, reference.lfns,
                                reference.nbfilestoget, reference.BkgEvtType, reference.ggtohadint,
                                reference.nbsigeventsperfile, reference.nbinputsigfile, reference.site,
                                reference.useEnergyForFileLookup, reference.machine, reference.pathToOverlayFiles,
                                ), self)
      if self.over.fcc is None:
        self.fail('FCC not initialized')
      if self.over.datMan is None:
        self.fail('DataManager not initialized')

  def test_applicationSpecificInputs_nodetector(self):
    self.over.detectormodel = ''
    assertDiracFailsWith(self.over.applicationSpecificInputs(),
                          'detector model not defined', self)

  def test_applicationSpecificInputs_noenergy(self):
    self.over.energytouse = ''
    assertDiracFailsWith(self.over.applicationSpecificInputs(),
                          'energy not set', self)

  def test_applicationSpecificInputs_nobxoverlay(self):
    self.over.BXOverlay = 0
    assertDiracFailsWith(self.over.applicationSpecificInputs(),
                          'bxoverlay parameter not defined', self)

  def test_applicationSpecificInputs_energyset_1(self):
    self.over.energytouse = ''
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=2)):
      self.over.energy = 123
      result = self.over.applicationSpecificInputs()
      assertDiracSucceedsWith_equals(result, TestOverlayUnittests.GOOD_EXIT, self)
      assertEqualsImproved(self.over.energytouse, '123gev', self)
      self.over.energy = 6800
      result = self.over.applicationSpecificInputs()
      assertDiracSucceedsWith_equals(result, TestOverlayUnittests.GOOD_EXIT, self)
      assertEqualsImproved(self.over.energytouse, '6.8tev', self)
      self.over.energy = 100000
      result = self.over.applicationSpecificInputs()
      assertDiracSucceedsWith_equals(result, TestOverlayUnittests.GOOD_EXIT, self)
      assertEqualsImproved(self.over.energytouse, '100tev', self)
      self.over.energy = 123.0
      result = self.over.applicationSpecificInputs()
      assertDiracSucceedsWith_equals(result, TestOverlayUnittests.GOOD_EXIT, self)
      assertEqualsImproved(self.over.energytouse, '123gev', self)

  def test_applicationSpecificInputs_with_setters(self):
    tmp_dict = {'Detector': 'othertestdetectorv3000', 'Energy': '10000GeV',
                 'BXOverlay': '651', 'ggtohadint': 9.5, 'ProdID': 429875,
                 'NbSigEvtsPerJob': 94, 'BkgEvtType': 'bgoijaf',
                 'STEP_NUMBER': 1,
               }
    self.over.step_commons = tmp_dict
    self.over.InputData = ['abc']
    self.over.NumberOfEvents = 15
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=2)):
      result = self.over.applicationSpecificInputs()
      assertDiracSucceedsWith_equals(result, TestOverlayUnittests.GOOD_EXIT, self)
      assertEqualsImproved((
          self.over.detectormodel, self.over.energytouse, self.over.BXOverlay,
          self.over.ggtohadint, self.over.prodid, self.over.NbSigEvtsPerJob,
          self.over.BkgEvtType), ('othertestdetectorv3000', '10000GeV', '651',
                                   9.5, 429875, 94, 'bgoijaf'), self)
      assertEqualsImproved(self.over.nbsigeventsperfile, 15, self)
      assertEqualsImproved(self.over.nbinputsigfile, 1, self)

  def test_applicationSpecificInputs_nonbevts(self):
    self.over.InputData = ['abc']
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=2)):
      result = self.over.applicationSpecificInputs()
      assertDiracFailsWith(result, 'number of events in the signal file is missing', self)

  def test_applicationSpecificInputs_nonbsigevts(self):
    self.over.NbSigEvtsPerJob = 0
    self.over.nbsigeventsperfile = 0
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=2)):
      result = self.over.applicationSpecificInputs()
      assertDiracFailsWith(result, 'could not determine the number of signal events per input file',
                            self)

  def test_applicationSpecificInputs_allowedBkg_rarepath(self):
    self.over.pathToOverlayFiles = 'some_path.txt'
    self.over.energy = self.over.energytouse = 0
    self.over.detector = self.over.detectormodel = ''
    result = self.over.applicationSpecificInputs()
    assertDiracSucceeds(result, self)

  def test_applicationSpecificInputs_allowedBkgFails_1(self):
    self.over.pathToOverlayFiles = 'some_path.txt'
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=-1)):
      result = self.over.applicationSpecificInputs()
      assertDiracFailsWith(result, 'no background to overlay', self)

  def test_applicationSpecificInputs_allowedBkgFails_2(self):
    self.over.pathToOverlayFiles = 'some_path.txt'
    self.over.detectormodel = ''
    self.over.detector = 'supergooddetectorv2'
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=-1)):
      result = self.over.applicationSpecificInputs()
      assertDiracFailsWith(result, 'no background to overlay', self)

  def test_getCastorFile(self):
    mylfn = '/ilc/user/j/jebbing/testfile.txt'
    expanded_lfn = '/castor/cern.ch/grid%s' % mylfn
    self.check_scriptwriting_method(mylfn, self.over.getCASTORFile, get_castor_lines(expanded_lfn))

  def test_getCastorFile_otherlfn(self):
    mylfn = '/castor/cern.ch/grid/ilc/user/j/jebbing/testfile.txt'
    expected = get_castor_lines(mylfn)
    expected[0].append("cp %s /tmp/x509up_u%s \n" % ('mytestproxy', 'mytestuserid'))
    self.check_scriptwriting_method(mylfn, self.over.getCASTORFile, expected,
                                     'failed', [False, False], False,
                                     environ_dict={'X509_USER_PROXY': 'mytestproxy'})

  def test_getLyonFile(self):
    mylfn = '/ilc/user/j/jebbing/testfile.txt'
    expanded_lfn = '/pnfs/in2p3.fr/data%s' % mylfn
    self.check_scriptwriting_method(mylfn, self.over.getLyonFile,
                                     get_lyon_lines(expanded_lfn),
                                     environ_dict={'X509_USER_PROXY': 'mytestproxy'})

  def test_getLyonFile_otherlfn(self):
    mylfn = '/pnfs/in2p3.fr/data/ilc/user/j/jebbing/testfile.txt'
    self.check_scriptwriting_method(mylfn, self.over.getLyonFile, get_lyon_lines(mylfn),
                                     'failed', [False, False], False,
                                     environ_dict={'X509_USER_PROXY': 'mytestproxy'})

  def test_getImperialFile(self):
    mylfn = '/ilc/user/j/jebbing/testfile.txt'
    expanded_lfn = '/pnfs/hep.ph.ic.ac.uk/data%s' % mylfn
    defaultse = 'defaultStorageElement_in_my_test'
    self.check_scriptwriting_method(mylfn, self.over.getImperialFile,
                                     get_imperial_lines(expanded_lfn, defaultse),
                                     environ_dict={'VO_ILC_DEFAULT_SE': defaultse})

  def test_getImperialFile_otherlfn(self):
    mylfn = '/pnfs/hep.ph.ic.ac.uk/data/ilc/user/j/jebbing/testfile.txt'
    defaultse = 'defaultStorageElement_in_my_test'
    self.check_scriptwriting_method(mylfn, self.over.getImperialFile,
                                     get_imperial_lines(mylfn, defaultse, True),
                                     'failed', [False, False, False], False, [[], []],
                                     [('overlayinput.sh', 'w'),
                                       (os.getcwd() + '/DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK', 'w')],
                                     {'VO_ILC_DEFAULT_SE': defaultse})

  def test_getRALFile(self):
    mylfn = '/ilc/user/j/jebbing/testfile.txt'
    expanded_lfn = '/castor/ads.rl.ac.uk/prod%s' % mylfn
    with patch('%s.subprocess.Popen' % MODULE_NAME, new=Mock()) as proc_mock:
      self.check_scriptwriting_method(mylfn, self.over.getRALFile, get_RAL_lines(expanded_lfn),
                                       environ_dict={'X509_USER_PROXY': 'mytestproxy'},
                                       is_ral=True)
      self.assertTrue(proc_mock.called)

  def test_getRALFile_otherlfn(self):
    mylfn = '/castor/ads.rl.ac.uk/prod/ilc/user/j/jebbing/testfile.txt'
    with patch('%s.subprocess.Popen' % MODULE_NAME, new=Mock()):
      self.check_scriptwriting_method(mylfn, self.over.getRALFile,
                                       get_RAL_lines(mylfn, True), 'failed', [False, False, False],
                                       False, [[], []], [
                                           ('overlayinput.sh', 'w'),
                                           (os.getcwd() + '/DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK', 'w')
                                           ], is_ral=True)

  def test_getKEKFile(self):
    mylfn = '/ilc/user/j/jebbing/testfile.txt'
    self.check_scriptwriting_method(mylfn, self.over.getKEKFile,
                                     get_KEK_lines('/grid%s' % mylfn),
                                     environ_dict={'X509_USER_PROXY': 'mytestproxy'})

  def test_getKEKFile_otherlfn(self):
    mylfn = '/ilc/user/j/jebbing/testfile.txt'
    self.check_scriptwriting_method(mylfn, self.over.getKEKFile,
                                     get_KEK_lines('/grid%s' % mylfn, True), 'failed',
                                     [False, False, False], False, [[], []],
                                     [('overlayinput.sh', 'w'),
                                       (os.getcwd() + '/DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK', 'w')])

  # pylint: disable=too-many-arguments
  def check_scriptwriting_method(self, mylfn, scriptmethod, expected, should_fail_with='',
                                  exists_sideeff=None, unlink_called=True,
                                  file_contents=None, expected_opens=None,
                                  environ_dict=None, is_ral=False):
    """Helper method that checks one of the methods provided by OverlayInput.

    :param str mylfn: file path (LFN)
    :param method scriptmethod: getLyonFile etc
    :param expected: Expected output in the script file (list of list of strings containing all touched files)
    :type expected: `python:list`
    :param str should_fail_with: Error message the method should return (in a S_ERROR structure)
    :param exists_sideeff: Return values of the os.path.exists method
    :type exists_sideeff: `python:list`
    :param bool unlink_called: indicates whether os.unlink is expected to be called or not
    :param expected_opens: a list of (filename, mode) pairs of all opened files
    :type expected_opens: `python:list`
    """
    # Set default list values
    if exists_sideeff is None:
      exists_sideeff = [True, True, True]
    if file_contents is None:
      file_contents = [[]]
    if environ_dict is None:
      environ_dict = {}
    if expected_opens is None:
      expected_opens = [('overlayinput.sh', 'w')]
    handles = FileUtil.getMultipleReadHandles(file_contents)
    # Variable mocks: exists return value, open values
    with patch('%s.shellCall' % MODULE_NAME, new=Mock(return_value=0)) as shell_mock, \
         patch('%s.os.unlink' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=exists_sideeff)), \
         patch('%s.open' % MODULE_NAME, mock_open(), create=True) as mo, \
         patch('%s.os.chmod' % MODULE_NAME, new=Mock(return_value=True)) as chmod_mock, \
         patch.dict(os.environ, environ_dict, True), \
         patch('%s.os.getuid' % MODULE_NAME, new=Mock(return_value='mytestuserid')):
      mo.side_effect = (h for h in handles)
      result = scriptmethod(mylfn)
      if should_fail_with:
        assertDiracFailsWith(result, should_fail_with, self)
      else:
        assertDiracSucceedsWith_equals(result, 'testfile.txt', self)
      shell_mock.assert_called_with(600, 'sh -c "./overlayinput.sh"',
                                     bufferLimit=20971520,
                                     callbackFunction=self.over.redirectLogOutput)
      chmod_mock.assert_called_with('overlayinput.sh', 0o755)
      if unlink_called:
        remove_mock.assert_called_with('overlayinput.sh')
      else:
        self.assertFalse(remove_mock.called)
      # Check if output to files is correct
      FileUtil.checkFileInteractions(self, mo, expected_opens, expected, handles)
      if is_ral:
        self.assertIn('CNS_HOST', os.environ)
        self.assertIn('STAGE_SVCCLASS', os.environ)
        self.assertIn('STAGE_HOST', os.environ)


class TestOverlayExecute(unittest.TestCase):
  """Tests the Execute method of the  Overlayinput class."""

  def setUp(self):
    self.over = OverlayInput()
    self.over.detectormodel = 'testdetectorv2000'
    self.over.energytouse = '200TeV'
    self.over.NbSigEvtsPerJob = 20
    self.over.pathToOverlayFiles = 'mytestfiles.txt'
    self.over.BXOverlay = 3
    self.over.ggtohadint = 5
    self.over.nbofeventsperfile = 21
    self.over.nbinputsigfile = 2
    self.over.site = "SomeSite"
    self.over.step_commons = dict(STEP_NUMBER=3)

  mockretval = S_OK({'Successful': {'testfile1.txt': ['CERN-DIP-4', 'KEK'],
                                      'testfile2.ppt': ['KEK']}, 'Failed': ''})

  def test_execute(self):
    rpc_mock = Mock()
    rpc_mock.canRun.return_value = S_OK(1)
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=2)), \
         patch('%s.FileCatalogClient.findFilesByMetadata' % MODULE_NAME, new=Mock(return_value=S_OK(['file1.txt', 'file2.ppt']))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
         patch('%s.open' % MODULE_NAME, mock_open(), create=True) as mo, \
         patch("%s.OverlaySystemClient" % MODULE_NAME, new=Mock(return_value=rpc_mock)), \
         patch('%s.os.mkdir' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.random.randrange' % MODULE_NAME, new=Mock(side_effect=[0, 0, 1, 1])), \
         patch('%s.wasteCPUCycles' % MODULE_NAME):
      self.over.getDataManagerFile = Mock(return_value=S_OK(True))
      result = self.over.execute()
      assertDiracSucceedsWith_equals(result, 'OverlayInput finished successfully', self)
      assertEqualsImproved(self.over.applicationLog, os.getcwd() + '/Overlay_input.log', self)

  def test_execute_resolve_fails(self):
    result = self.over.execute()
    assertDiracFailsWith(result, 'no background to overlay', self)

  def test_execute_status_not_ok(self):
    log = 'my_123_log.txt'
    self.over.applicationLog = log
    self.over.workflowStatus = S_ERROR('myerror167')
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=2)):
      result = self.over.execute()
      assertDiracSucceedsWith(result, 'OverlayInput should not proceed', self)
      assertEqualsImproved(self.over.applicationLog, os.getcwd() + '/' + log, self)

  def test_execute_getfiles_fails(self):
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=2)), \
        patch('%s.OverlayInput._OverlayInput__getFilesFromPath' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_getfile_error'))):
      assertDiracFailsWith(self.over.execute(), 'some_getfile_error', self)

  def test_execute_getfiles_empty(self):
    self.over.pathToOverlayFiles = ''
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=2)), \
         patch('%s.OverlayInput._OverlayInput__getFilesFromFC' % MODULE_NAME, new=Mock(return_value=S_OK([]))):
      assertDiracFailsWith(self.over.execute(), 'overlayprocessor got an empty list', self)

  def test_execute_getlocally_fails(self):
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=2)), \
        patch('%s.OverlayInput._OverlayInput__getFilesFromPath' % MODULE_NAME, new=Mock(return_value=S_OK(['mylfn1', 'otherlfn', 'many_more_lfns.txt']))), \
        patch('%s.OverlayInput._OverlayInput__getFilesLocaly' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_local_getfile_err'))):
      assertDiracFailsWith(self.over.execute(), 'failed to get files locally', self)

  # pylint: disable=protected-access,no-member
  def test_getfcfiles(self):
    ops_dict = {'/Overlay/clic_cdr/200TeV/testdetectorv2000/myTestBkgEvt/ProdID': 98421,
                 '/Overlay/clic_cdr/200TeV/testdetectorv2000/myTestBkgEvt/NbEvts': 482,
                 '/Overlay/clic_cdr/200TeV/testdetectorv2000/myTestBkgEvt/EvtType': 'someTestEventType'}
    self.over.energy = 123
    self.over.useEnergyForFileLookup = True
    self.over.BkgEvtType = 'myTestBkgEvt'
    self.over.machine = 'clic_cdr'
    ops_mock = Mock()
    ops_mock.getValue.side_effect = lambda key, default: ops_dict[key]
    self.over.ops = ops_mock
    fcc_mock = Mock()
    fcc_mock.findFilesByMetadata.return_value = S_OK(9824)
    self.over.fcc = fcc_mock
    result = self.over._OverlayInput__getFilesFromFC()
    assertDiracSucceedsWith_equals(result, 9824, self)
    fcc_mock.findFilesByMetadata.assert_called_once_with(
        {'Energy': '123', 'EvtType': 'someTestEventType', 'ProdID': 98421, 'Datatype': 'SIM',
         'DetectorModel': 'testdetectorv2000', 'Machine': 'clic'})

  def test_getfcfiles_othercase(self):
    ops_dict = {'/Overlay/ilc_dbd/TestILCDetectorv1/200TeV/otherTestEvt/ProdID': 139,
                 '/Overlay/ilc_dbd/200TeV/TestILCDetectorv1/otherTestEvt/NbEvts': 2145,
                 '/Overlay/ilc_dbd/200TeV/TestILCDetectorv1/otherTestEvt/EvtType': 'ilc_evt_testme'}
    self.over.energy = 0
    self.over.useEnergyForFileLookup = False
    self.over.detectormodel = ''
    self.over.BkgEvtType = 'otherTestEvt'
    self.over.machine = 'ilc_dbd'
    self.over.detector = 'TestILCDetectorv1'
    self.over.prodid = 82492
    ops_mock = Mock()
    ops_mock.getValue.side_effect = lambda key, default: ops_dict[key]
    self.over.ops = ops_mock
    fcc_mock = Mock()
    fcc_mock.findFilesByMetadata.return_value = S_OK(2948)
    self.over.fcc = fcc_mock
    result = self.over._OverlayInput__getFilesFromFC()
    assertDiracSucceedsWith_equals(result, 2948, self)
    fcc_mock.findFilesByMetadata.assert_called_once_with(
        {'EvtType': 'ilc_evt_testme', 'ProdID': 82492, 'Datatype': 'SIM', 'Machine': 'ilc'})

  def test_getfilesfromFC(self):
    ops_mock = Mock()
    ops_mock.getValue.side_effect = ['1245', 2849, None]
    self.over.ops = ops_mock
    fcc_mock = Mock()
    fcc_mock.findFilesByMetadata.return_value = S_OK('1245')
    self.over.fcc = fcc_mock
    self.over.energy = 9842
    self.over.useEnergyForFileLookup = True
    self.over.detectormodel = 'myTestDetectorv021'
    self.over.machine = 'clic_cdr'
    self.over.detector = 'overlaydetector'
    assertDiracSucceeds(self.over._OverlayInput__getFilesFromFC(), self)
    fcc_mock.findFilesByMetadata.assert_called_once_with(
        {'Energy': '9842', 'EvtType': None, 'Datatype': 'SIM', 'DetectorModel': 'myTestDetectorv021',
         'Machine': 'clic', 'ProdID': '1245'})

  def test_getfilesfromlyon(self):
    import subprocess
    popen_mock = Mock()
    popen_mock.communicate.side_effect = [('myfile1\nfile1923\n813tev_collision  ', 'ignoreme'),
                                           ('file1.stdhep\nsome_other_file', 'ignoreme'), ('', 'ignoreme'),
                                           ('\nlast_file.txt', 'ignoreme')]
    lyon_dict = {'ProdID': 121345, 'Energy': '813', 'EvtType': 'myTestEvt',
                  'DetectorType': 'myTestDetectorv3'}
    with patch('subprocess.Popen', new=Mock(return_value=popen_mock)) as proc_mock:
      proc_command_dir = '/ilc/prod/clic/813/myTestEvt/myTestDetectorv3/SIM/00121345/'
      assertDiracSucceedsWith_equals(self.over._OverlayInput__getFilesFromLyon(lyon_dict),
                                      [proc_command_dir + 'myfile1/file1.stdhep',
                                        proc_command_dir + 'myfile1/some_other_file',
                                        proc_command_dir + 'file1923/', proc_command_dir + '813tev_collision/',
                                        proc_command_dir + '813tev_collision/last_file.txt'], self)
      assertEqualsImproved(proc_mock.mock_calls,
                            [call(['nsls', proc_command_dir], stdout=subprocess.PIPE),
                              call(['nsls', proc_command_dir + 'myfile1'], stdout=subprocess.PIPE),
                              call(['nsls', proc_command_dir + 'file1923'], stdout=subprocess.PIPE),
                              call(['nsls', proc_command_dir + '813tev_collision'],
                                    stdout=subprocess.PIPE)], self)  # Checks for expected calls on the mock

  def test_getfilesfromlyon_ignore_all(self):
    import subprocess
    popen_mock = Mock()
    popen_mock.communicate.side_effect = [('mypaths/dirac_directory/some/other/stuff\nmy/dir ', 'ignoreme'),
                                           ('other_file/dirac_directory', 'ignoreme')]
    lyon_dict = {'ProdID': 121345, 'Energy': '813', 'EvtType': 'myTestEvt',
                  'DetectorType': 'myTestDetectorv3'}
    with patch('subprocess.Popen', new=Mock(return_value=popen_mock)) as proc_mock:
      proc_command_dir = '/ilc/prod/clic/813/myTestEvt/myTestDetectorv3/SIM/00121345/'
      assertDiracFailsWith(self.over._OverlayInput__getFilesFromLyon(lyon_dict), 'file list is empty',
                            self)
      assertEqualsImproved(proc_mock.mock_calls,
                            [call(['nsls', proc_command_dir], stdout=subprocess.PIPE),
                              call(['nsls', proc_command_dir + 'my/dir'], stdout=subprocess.PIPE)], self)

  def test_getfilesfromcastor(self):
    self.over.machine = 'testMach12'
    import subprocess
    popen_mock = Mock()
    popen_mock.communicate.side_effect = [('myfile1\nfile1923\n813tev_collision  ', 'ignoreme'),
                                           ('file1.stdhep\nsome_other_file', 'ignoreme'), ('', 'ignoreme'),
                                           ('\nlast_file.txt', 'ignoreme')]
    castor_dict = {'ProdID': 121345, 'Energy': '813', 'EvtType': 'myTestEvt',
                    'DetectorType': 'myTestDetectorv3'}
    with patch('subprocess.Popen', new=Mock(return_value=popen_mock)) as proc_mock:
      proc_command_dir = '/castor/cern.ch/grid/ilc/prod/testMach12/813/myTestEvt/myTestDetectorv3/SIM/00121345/'
      assertDiracSucceedsWith_equals(self.over._OverlayInput__getFilesFromCastor(castor_dict),
                                      [proc_command_dir + 'myfile1/file1.stdhep',
                                        proc_command_dir + 'myfile1/some_other_file',
                                        proc_command_dir + 'file1923/', proc_command_dir + '813tev_collision/',
                                        proc_command_dir + '813tev_collision/last_file.txt'], self)
      assertEqualsImproved(proc_mock.mock_calls,
                            [call(['nsls', proc_command_dir], stdout=subprocess.PIPE),
                              call(['nsls', proc_command_dir + 'myfile1'], stdout=subprocess.PIPE),
                              call(['nsls', proc_command_dir + 'file1923'], stdout=subprocess.PIPE),
                              call(['nsls', proc_command_dir + '813tev_collision'],
                                    stdout=subprocess.PIPE)], self)

  def test_getfilesfromcastor_ignoreall(self):
    import subprocess
    popen_mock = Mock()
    popen_mock.communicate.side_effect = [('mypaths/dirac_directory/some/other/stuff\nmy/dir ', 'ignoreme'),
                                           ('other_file/dirac_directory', 'ignoreme')]
    castor_dict = {'ProdID': 121345, 'Energy': '813', 'EvtType': 'myTestEvt',
                    'DetectorType': 'myTestDetectorv3'}
    with patch('subprocess.Popen', new=Mock(return_value=popen_mock)) as proc_mock:
      proc_command_dir = '/castor/cern.ch/grid/ilc/prod/clic_cdr/813/myTestEvt/myTestDetectorv3/SIM/00121345/'
      assertDiracFailsWith(self.over._OverlayInput__getFilesFromCastor(castor_dict), 'file list is empty',
                            self)
      assertEqualsImproved(proc_mock.mock_calls,
                            [call(['nsls', proc_command_dir], stdout=subprocess.PIPE),
                              call(['nsls', proc_command_dir + 'my/dir'], stdout=subprocess.PIPE)], self)

  @parameterized.expand([(True, {}),
                         (False, {'eventsPerFile': 1}),
                         (True, {'sigPerJob': 1}),
                         (False, {'sigPerJob': None}),
                         (True, {'sigPerJob': None, 'neSigFile': 10, 'neSigPerFile': 100}),
                         (False, {'overlayrun': False}),
                         (False, {'overlayrun': True, 'orc': 0}),
                         (False, {'datManGet': False, 'ERROR': 'Failed to get files'}),
                         ])
  def test_getFilesLocaly(self, success, pars):
    """Test getFilesLocaly function."""
    rpc_mock = Mock()
    rpc_mock.canRun.return_value = S_OK(pars.get('orc', 1)) if pars.get('overlayrun', True) else S_ERROR('nope')
    dms = pars.get('datManGet', True)
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value=2)), \
            patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)), \
            patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)), \
            patch('%s.open' % MODULE_NAME, mock_open(), create=True), \
            patch("%s.OverlaySystemClient" % MODULE_NAME, new=Mock(return_value=rpc_mock)), \
            patch('%s.os.mkdir' % MODULE_NAME, new=Mock(return_value=True)), \
            patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)), \
            patch('%s.time.sleep' % MODULE_NAME, new=Mock(return_value=True)), \
            patch('%s.random.randrange' % MODULE_NAME, new=Mock(side_effect=[0, 0, 1, 1, 1, 1, 1, 2, 2, 2])), \
            patch('%s.wasteCPUCycles' % MODULE_NAME):
      self.over.BXOverlay = 20
      self.over.ggtohadint = 3
      self.over.nbofeventsperfile = pars.get('eventsPerFile', 100)
      self.over.NbSigEvtsPerJob = pars.get('sigPerJob', 100)
      self.over.nbinputsigfile = pars.get('neSigFile', 0)
      self.over.nbsigeventsperfile = pars.get('neSigPerFile', 0)
      self.over.lfns = ['foo', 'bar', 'baz']
      self.over.getDataManagerFile = Mock(return_value=S_OK(True) if dms else S_ERROR('nope'))
      result = self.over._OverlayInput__getFilesLocaly()
      assert result['OK'] == success, result.get('Message', 'No error')
      if pars.get('ERROR'):
        assert result['Message'] == pars.get('ERROR')

  @parameterized.expand([(True, {}),
                         (False, {'acFailure': 'Activity Failed'}),
                         (True, {'dmFail': True}),
                         (False, {'dmFullFail': True}),
                         ])
  def test_getDataManagerFile(self, success, pars):
    """Test getDataManagerFile function."""
    lfn = '/vo/foo/bar'
    dMock = Mock()
    self.over.datMan = dMock
    dMock.getFile = Mock(return_value=S_OK({'Successful': {lfn: True}, 'Failed': {}}))
    dMock.getActiveReplicas = Mock(return_value=S_OK({'Successful': {lfn: {'FOO-FAR': True,
                                                                           'FOO-FAR-FAR': True,
                                                                           'FOO-FAR-FAR-FAR': True,
                                                                           'FOO-CLOSE': True}}}))
    if pars.get('acFailure'):
      dMock.getActiveReplicas = Mock(return_value=S_ERROR(pars.get('acFailure')))
    if pars.get('dmFail'):
      dMock.getFile = Mock(side_effect=[S_OK({'Successful': {}, 'Failed': {lfn: True}}),
                                        S_OK({'Successful': {lfn: True}, 'Failed': {}})])
    if pars.get('dmFullFail'):
      dMock.getFile = Mock(return_value=S_ERROR('fail'))

    def distance(SE):
      return int(SE.count('FAR'))
    self.over._OverlayInput__distanceToSE = Mock(side_effect=distance)

    res = self.over.getDataManagerFile(lfn)
    assert res['OK'] == success
    if not pars.get('acFailure'):
      dMock.getFile.assert_any_call(lfn, sourceSE='FOO-CLOSE')
    if pars.get('dmFail'):
      dMock.getFile.assert_any_call(lfn, sourceSE='FOO-CLOSE')
      dMock.getFile.assert_called_with(lfn, sourceSE='FOO-FAR')
    if pars.get('dmFullFail'):
      dMock.getFile.assert_any_call(lfn, sourceSE='FOO-FAR')
      dMock.getFile.assert_any_call(lfn, sourceSE='FOO-FAR-FAR')
      dMock.getFile.assert_called_with(lfn, sourceSE='FOO-FAR-FAR-FAR')
      assert res['Message'].endswith('any sourceSE')

    if pars.get('acFailure'):
      assert res['Message'] == pars.get('acFailure')

  @parameterized.expand([(True, {'site': 'LCG.baz.ch', 'distance': 0}),
                         (True, {'site': 'LCG.bar.pu', 'distance': 13}),
                         ])
  def test_distanceToSE(self, success, pars):
    """Test __distanceToSE function."""
    storageMock = [Mock(name='TheStorage')]
    storageMock[0].getParameters = Mock(return_value={'Host': 'se.baz.ch'})
    seInstance = Mock(name='SE Instance')
    seInstance.storages = storageMock
    seMock = Mock(return_value=seInstance, name='SE Class')
    cMock = Mock(name='gConfig')
    sites = ['LCG.bar.pu', 'LCG.baz.ch']

    def cmg(path, default):
      """Mock for config getValue."""
      print("Calling cmg with", path, default)
      if '.pu' in path:
        return '90:90'
      if '.ch' in path:
        return '20:20'
      return '50:50'
    cMock.getValue.side_effect = cmg
    self.over.site = pars.get('site', 'LCG.baz.ch')
    with patch('%s.StorageElement' % MODULE_NAME, new=seMock), \
            patch('%s.gConfig' % MODULE_NAME, new=cMock), \
            patch('%s.getSitePath' % MODULE_NAME, new=Mock(side_effect=lambda x: S_OK('/path/' + x))), \
            patch('%s.getSites' % MODULE_NAME, new=Mock(return_value=S_OK(sites))):
      distance = self.over._OverlayInput__distanceToSE('FOO-SE')
    if success:
      assert int(distance * 10) == pars.get('distance', -1)

  @parameterized.expand([(True, True, [-180, -90]),
                         (False, True, [0, -90]),
                         (False, False, [0.0, 0.0]),
                         ])
  def test_getSELocation(self, paraFails, siteFails, expected):
    """Test getSELocation function."""
    storageMock = [Mock(name='TheStorage')]
    storageMock[0].getParameters = Mock(return_value={'NoHost': None} if paraFails else {'Host': 'par.ty'})
    seInstance = Mock(name='SE Instance')
    seInstance.storages = storageMock
    seMock = Mock(return_value=seInstance, name='SE Class')
    with patch('%s.StorageElement' % MODULE_NAME, new=seMock), \
            patch('%s.getSites' % MODULE_NAME, new=Mock(return_value=S_ERROR('nope') if siteFails else S_OK([]))):
      assert self.over._OverlayInput__getSELocation('FOO-SE') == expected

  @parameterized.expand([(True, [0.0, 0.0]),
                         (False, [50, 50]),
                         ])
  def test_getSiteLocation(self, sitePathFails, expected):
    """Test getSiteLocation function."""
    def cmg(path, default):
      return '50:50'
    cMock = Mock(name='gConfig')
    cMock.getValue.side_effect = cmg
    gspMock = Mock()
    if sitePathFails:
      gspMock.return_value = S_ERROR('nope')
    else:
      gspMock.side_effect = lambda x: S_OK('/path/' + x)
    with patch('%s.getSitePath' % MODULE_NAME, new=gspMock), \
            patch('%s.gConfig' % MODULE_NAME, new=cMock):
      assert self.over._OverlayInput__getSiteLocation(site='aSite') == expected


def get_castor_lines(expanded_lfn):
  result = [[
      '#!/bin/sh \n', '###############################\n', '# Dynamically generated scrip #\n',
      '###############################\n', 'declare -x STAGE_SVCCLASS=ilcdata\n',
      'declare -x STAGE_HOST=castorpublic\n',
      r"xrdcp -s root://castorpublic.cern.ch/%s ./ -OSstagerHost=castorpublic\&svcClass=ilcdata\n" % expanded_lfn,
      """
if [ ! -s %s ]; then
  echo "Using rfcp instead"
  rfcp %s ./
fi\n""" % ('testfile.txt', expanded_lfn), 'declare -x appstatus=$?\n',
      'exit $appstatus\n']]
  return result


def get_lyon_lines(expanded_lfn):
  result = [[
      '#!/bin/sh \n', '###############################\n', '# Dynamically generated scrip #\n',
      '###############################\n', "cp %s /tmp/x509up_u%s \n" % ('mytestproxy', 'mytestuserid'),
      ". /afs/in2p3.fr/grid/profiles/lcg_env.sh\n",
      "xrdcp root://ccdcacsn179.in2p3.fr:1094%s ./ -s\n" % expanded_lfn,
      'declare -x appstatus=$?\n', 'exit $appstatus\n']]
  return result


def get_imperial_lines(expanded_lfn, defaultse, with_watchdog=False):
  result = []
  if with_watchdog:
    result.append(['Dont look at cpu'])
  result.append(['#!/bin/sh \n', '###############################\n',
                   '# Dynamically generated scrip #\n', '###############################\n',
                   "dccp dcap://%s%s ./\n" % (defaultse, expanded_lfn),
                   'declare -x appstatus=$?\n', 'exit $appstatus\n'])
  return result


def get_RAL_lines(expanded_lfn, with_watchdog=False):
  result = []
  if with_watchdog:
    result.append(['Dont look at cpu'])
  result.append(['#!/bin/sh \n',
                 '###############################\n',
                 '# Dynamically generated scrip #\n',
                 '###############################\n',
                 "/usr/bin/rfcp 'rfio://cgenstager.ads.rl.ac.uk:9002?svcClass=ilcTape&path=%s' %s\n" % (expanded_lfn,
                                                                                                        'testfile.txt'),
                 'declare -x appstatus=$?\n',
                 'exit $appstatus\n'])
  return result


def get_KEK_lines(expanded_lfn, with_watchdog=False):
  result = []
  if with_watchdog:
    result.append(['Dont look at cpu'])
  result.append(['#!/bin/sh \n', '###############################\n', '# Dynamically generated scrip #\n',
                   '###############################\n', "cp %s ./ -s\n" % expanded_lfn,
                   'declare -x appstatus=$?\n', 'exit $appstatus\n'])
  return result


def runTests():
  """Runs our tests."""
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestOverlayEos)

  testResult = unittest.TextTestRunner(verbosity=2).run(suite)
  print(testResult)

  suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestOverlayUnittests)

  testResult = unittest.TextTestRunner(verbosity=2).run(suite)
  print(testResult)

  suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestOverlayExecute)

  testResult = unittest.TextTestRunner(verbosity=2).run(suite)
  print(testResult)


if __name__ == '__main__':
  runTests()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

########################################################################
# File :   InputDataResolution.py
# Author : S Poss, based on S Paterson
########################################################################

"""The input data resolution module is a VO-specific plugin that allows to define VO input data policy in a simple way using existing utilities in DIRAC or extension code supplied by the VO.

The arguments dictionary from the Job Wrapper includes the file catalogue
result and in principle has all necessary information to resolve input data
for applications.
"""

from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.ModuleFactory import ModuleFactory
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

from DIRAC import S_OK, S_ERROR, gLogger
import DIRAC


LOG = gLogger.getSubLogger(__name__)
COMPONENT_NAME = 'ILCInputDataResolution'


class InputDataResolution(object):
  """ILC specific input data resolution, imported from DIRAC."""
  #############################################################################

  def __init__(self, argumentsDict):
    """Standard constructor."""
    self.arguments = argumentsDict
    self.name = COMPONENT_NAME
    self.ops = Operations()
  #############################################################################

  def execute(self):
    """Given the arguments from the Job Wrapper, this function calls existing utilities in DIRAC to resolve input data according to LHCb VO policy."""
    result = self.__resolveInputData()
    if not result['OK']:
      LOG.error('InputData resolution failed with result:\n%s' % (result))
      LOG.error('on no, double error')

    # For local running of this module we can expose an option to ignore missing files
    ignoreMissing = False
    if 'IgnoreMissing' in self.arguments:
      ignoreMissing = self.arguments['IgnoreMissing']

    # For LHCb original policy was as long as one TURL exists, this can be conveyed to the application
    # this breaks due to the stripping so the policy has been changed.
    if 'Failed' in result:
      failedReplicas = result['Failed']
      if failedReplicas and not ignoreMissing:
        LOG.error('Failed to obtain access to the following files:\n%s' % ('\n'.join(failedReplicas)))
        return S_ERROR('Failed to access all of requested input data')

    if 'Successful' not in result:
      return result

    if not result['Successful']:
      return S_ERROR('Could not access any requested input data')

    return result

  #############################################################################
  def __resolveInputData(self):
    """This method controls the execution of the DIRAC input data modules according to the ILC VO policy defined in the configuration service."""
    if 'SiteName' in self.arguments['Configuration']:
      site = self.arguments['Configuration']['SiteName']
    else:
      site = DIRAC.siteName()

    policy = []
    if 'Job' not in self.arguments:
      self.arguments['Job'] = {}

    if 'InputDataPolicy' in self.arguments['Job']:
      policy = self.arguments['Job']['InputDataPolicy']
      # In principle this can be a list of modules with the first taking precedence
      if type(policy) in (str,):
        policy = [policy]
      LOG.info('Job has a specific policy setting: %s' % (', '.join(policy)))
    else:
      LOG.verbose('Attempting to resolve input data policy for site %s' % site)
      inputDataPolicy = self.ops.getOptionsDict('/InputDataPolicy')
      if not inputDataPolicy:
        return S_ERROR('Could not resolve InputDataPolicy from /InputDataPolicy')

      options = inputDataPolicy['Value']
      if site in options:
        policy = options[site]
        policy = [x.strip() for x in policy.split(',')]
        LOG.info('Found specific input data policy for site %s:\n%s' % (site, ',\n'.join(policy)))
      elif 'Default' in options:
        policy = options['Default']
        policy = [x.strip() for x in policy.split(',')]
        LOG.info('Applying default input data policy for site %s:\n%s' % (site, ',\n'.join(policy)))

    dataToResolve = None  # if none, all supplied input data is resolved
    allDataResolved = False
    successful = {}
    failedReplicas = []
    for modulePath in policy:
      if not allDataResolved:
        result = self.__runModule(modulePath, dataToResolve)
        if not result['OK']:
          LOG.warn('Problem during %s execution' % modulePath)
          return result

        if 'Failed' in result:
          failedReplicas = result['Failed']

        if failedReplicas:
          LOG.info('%s failed for the following files:\n%s' % (modulePath, '\n'.join(failedReplicas)))
          dataToResolve = failedReplicas
        else:
          LOG.info('All replicas resolved after %s execution' % (modulePath))
          allDataResolved = True

        successful.update(result['Successful'])
        LOG.verbose(successful)

    result = S_OK()
    result['Successful'] = successful
    result['Failed'] = failedReplicas
    return result

  #############################################################################
  def __runModule(self, modulePath, remainingReplicas):
    """This method provides a way to run the modules specified by the VO that govern the input data access policy for the current site.

    For LHCb the
    standard WMS modules are applied in a different order depending on the site.
    """
    LOG.info('Attempting to run %s' % (modulePath))
    moduleFactory = ModuleFactory()
    moduleInstance = moduleFactory.getModule(modulePath, self.arguments)
    if not moduleInstance['OK']:
      return moduleInstance

    module = moduleInstance['Value']
    result = module.execute(remainingReplicas)
    return result

#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Add a number of tasks to a production, can only be used on productions that use the 'Limited' or 'Standard' Plugin.

Options:

   -p, --ProductionID <value>   Production ID to extend. Deprecated: use positional args
   -t, --Tasks <value>          Number of tasks to add (-1 for all) Deprecated: use positional args
     , --Total                  Set MaxNumberOfTasks to the value of TasksToAdd

Usage:

  dirac-ilc-add-tasks-to-prod ProdID TasksToAdd [--Total]


:since: Mar 26, 2013
:author: S Poss
"""

from __future__ import absolute_import
import textwrap

from DIRAC.Core.Base import Script
from DIRAC.Core.Utilities.DIRACScript import DIRACScript
from DIRAC import S_OK, S_ERROR
from DIRAC import gLogger, exit as dexit


__RCSID__ = '$Id$'


class _Skip(Exception):
  """Exception to skip further execution of script."""

  pass


class _Params(object):
  def __init__(self):
    self.prod = 0
    self.tasks = 0
    self.total = False

  def setProd(self, opt):
    try:
      self.prod = int(opt)
    except ValueError:
      return S_ERROR('Production ID must be integer')
    return S_OK()

  def setNbTasks(self, opt):
    try:
      self.tasks = int(opt)
    except ValueError:
      return S_ERROR('Number of tasks must be integer')
    return S_OK()

  def setTotal(self, _opt):
    self.total = True
    return S_OK()

  def registerSwitches(self):
    Script.registerSwitch('p:', 'ProductionID=', 'Production ID to extend', self.setProd)
    Script.registerSwitch('t:', 'Tasks=', 'Number of tasks to add (-1 for all)', self.setNbTasks)
    Script.registerSwitch('', 'Total', 'Set MaxNumberOfTasks to the value of TasksToAdd', self.setTotal)
    Script.setUsageMessage(textwrap.dedent("""
                           Usage:

                             %s 2145 200 [--Total]

                           Deprecated Usage:

                             %s -p 2145 -t 200 [--Total]

                           """ % (Script.scriptName, Script.scriptName)))

  def getMaxTasks(self, maxTasks):
    """Return the new max number of tasks."""
    if self.total:
      if self.tasks > maxTasks:
        return self.tasks
      gLogger.notice('MaxNumberOfTasks is already reached for production %s' % self.prod)
      raise _Skip()
    # if not total, add number of tasks to current maxTasks
    gLogger.notice('Adding %s tasks to production %s' % (self.tasks, self.prod))
    return maxTasks + self.tasks


def _getTransformationClient():
  """Return instance of TransformationClient."""
  from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
  return TransformationClient()


def _extendLimited(clip, tc, trans):
  """Extend transformations with 'Limited' plugin."""
  if clip.tasks > 0:
    max_tasks = clip.getMaxTasks(trans['MaxNumberOfTasks'])
    groupsize = trans['GroupSize']
    gLogger.notice('Adding %s tasks (%s file(s)) to production %s' % (clip.tasks, clip.tasks * groupsize, clip.prod))
  elif clip.tasks < 0:
    max_tasks = -1
    gLogger.notice('Now all existing files in the DB for production %s will be processed.' % clip.prod)
  else:
    gLogger.error('Number of tasks must be different from 0')
    return 1

  res = tc.setTransformationParameter(clip.prod, 'MaxNumberOfTasks', max_tasks)
  if not res['OK']:
    gLogger.error('Failed to set MaxNumberOfTasks', res['Message'])
    return 1
  gLogger.notice('Production %s extended!' % clip.prod)
  return 0


def _extendStandard(clip, tc, trans):
  """Extend transformations with 'Standard' Plugin."""
  if clip.tasks > 0:
    max_tasks = clip.getMaxTasks(trans['MaxNumberOfTasks'])
  else:
    gLogger.error('Number of tasks must be larger than 0')
    return 1

  res = tc.setTransformationParameter(clip.prod, 'MaxNumberOfTasks', max_tasks)
  if not res['OK']:
    gLogger.error('Failed to set MaxNumberOfTasks', res['Message'])
    return 1

  gLogger.notice('Production %s extended!' % clip.prod)
  return 0


def _extend(clip, posArgs):
  """Extend the production for given number of tasks."""

  if posArgs and (clip.prod or clip.tasks):
    gLogger.error('Do not mix positional and keyword arguments')
    return 1

  if len(posArgs) == 2:
    clip.setProd(posArgs[0])
    clip.setNbTasks(posArgs[1])

  if not (clip.prod or clip.tasks):
    gLogger.error('Production ID is 0 or Tasks is 0, cannot be')
    return 1

  tc = _getTransformationClient()
  res = tc.getTransformation(clip.prod)
  trans = res['Value']
  transp = trans['Plugin']
  gLogger.info('Production "%s" has "%s" tasks registered' % (clip.prod, trans['MaxNumberOfTasks']))

  if not posArgs:
    gLogger.notice('You can also just use positional arguments now!')

  if transp == 'Limited':
    return _extendLimited(clip, tc, trans)

  if transp == 'Standard':
    return _extendStandard(clip, tc, trans)

  gLogger.error("This cannot be used on productions that are not using the 'Limited' or 'Standard' plugin")
  return 1


@DIRACScript()
def main():
  CLIP = _Params()
  CLIP.registerSwitches()
  Script.parseCommandLine()
  POS_ARGS = Script.getPositionalArgs()
  try:
    dexit(_extend(CLIP, POS_ARGS))
  except _Skip:
    gLogger.verbose('Skipping extension')
    dexit(1)


if __name__ == "__main__":
  main()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for JobResetAgent."""
# pylint: disable=protected-access
from __future__ import absolute_import
from contextlib import contextmanager
from datetime import datetime, timedelta

import sys
import pytest
from mock import MagicMock, call

import ILCDIRAC.WorkloadManagementSystem.Agent.JobResetAgent as JRA

import DIRAC.Resources.Storage.StorageElement as SeModule
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation

from DIRAC import S_OK, S_ERROR
from DIRAC import gLogger
import DIRAC

__RCSID__ = "$Id$"


@pytest.fixture
def userProxyFixture(mocker):
  """Mock UserProxy."""
  @contextmanager
  def _mockedCM(*args, **kwargs):
    try:
      yield S_OK()
    finally:
      pass
  mocker.patch('ILCDIRAC.WorkloadManagementSystem.Agent.JobResetAgent.UserProxy', new=_mockedCM)


@pytest.fixture
def failingUserProxyFixture(mocker):
  """Mock UserProxy."""
  @contextmanager
  def _mockedCM(*args, **kwargs):
    try:
      yield S_ERROR("Failed to set up proxy")
    finally:
      pass
  mocker.patch('ILCDIRAC.WorkloadManagementSystem.Agent.JobResetAgent.UserProxy', new=_mockedCM)


@pytest.fixture
def fakeJobID():
  """Return fake job ID."""
  return 1


@pytest.fixture
def alreadyErrorMessage():
  """Return error message."""
  return "File already registered with alternative metadata"


@pytest.fixture
def today():
  """Return a day."""
  return datetime(2018, 12, 25, 0, 0, 0, 0)


@pytest.fixture
def jobResetAgent(today):
  """Fixture for jobResetAgent."""
  import ILCDIRAC.WorkloadManagementSystem.Agent.JobResetAgent as JRA
  from ILCDIRAC.WorkloadManagementSystem.Agent.JobResetAgent import JobResetAgent
  agent = JRA
  agent.AgentModule = MagicMock()
  agent.JobMonitoringClient = MagicMock()
  agent.DataManager = MagicMock(spec=DIRAC.DataManagementSystem.Client.DataManager.DataManager)
  agent.ReqClient = MagicMock(spec=DIRAC.RequestManagementSystem.Client.ReqClient.ReqClient)
  agent.NotificationClient = MagicMock(spec=DIRAC.FrameworkSystem.Client.NotificationClient.NotificationClient)
  agent.datetime = MagicMock()
  agent.datetime.now.return_value = today

  theAgent = JobResetAgent()
  theAgent.log = gLogger
  theAgent.enabled = True
  theAgent._fcClient = MagicMock(spec=DIRAC.Resources.Catalog.FileCatalogFactory.FileCatalogFactory)
  theAgent._fcClient.removeFile = MagicMock(return_value=S_ERROR('failed'))
  theAgent.jobManagerClient = MagicMock()
  theAgent.jobStateUpdateClient = MagicMock()
  theAgent._jobDB = MagicMock()
  yield theAgent
  sys.modules.pop('ILCDIRAC.WorkloadManagementSystem.Agent.JobResetAgent')


def createRequest(requestID, opType, opStatus, fileStatus, lfnError=" ",
                  lfn="/ilc/fake/lfn"):
  """Create a request."""
  req = Request({"RequestID": requestID})
  op = Operation({"Type": opType, "Status": opStatus})
  op.addFile(File({"LFN": lfn, "Status": fileStatus, "Error": lfnError}))
  req.addOperation(op)
  return req


@pytest.fixture
def doneRemoveRequest():
  """Return done Remove Request."""
  return createRequest(requestID=1, opType="RemoveFile",
                       opStatus="Done", fileStatus="Done")


@pytest.fixture
def doneReplicateRequest():
  """Return done Replicate Request."""
  return createRequest(requestID=2, opType="ReplicateAndRegister",
                       opStatus="Done", fileStatus="Done")


@pytest.fixture
def failedReplicateRequest():
  """Return Failed Replicate Request."""
  return createRequest(requestID=3, opType="ReplicateAndRegister",
                       opStatus="Failed", fileStatus="Failed")


@pytest.fixture
def failedRemoveRequest():
  """Return Failed RemoveRequest Request."""
  return createRequest(requestID=4, opType="RemoveFile",
                       opStatus="Failed", fileStatus="Failed")


@pytest.fixture
def failedRemoveReplicaRequest():
  """Return failed RemoveReplica Request."""
  return createRequest(requestID=5, opType='RemoveReplica',
                       opStatus='Failed', fileStatus='Failed')


def test_init(jobResetAgent):
  """Test the constructor."""
  from ILCDIRAC.WorkloadManagementSystem.Agent.JobResetAgent import JobResetAgent
  assert isinstance(jobResetAgent, JobResetAgent)
  assert isinstance(jobResetAgent.jobMonClient, MagicMock)
  assert isinstance(jobResetAgent.dataManager, MagicMock)
  assert isinstance(jobResetAgent.reqClient, MagicMock)
  assert isinstance(jobResetAgent.nClient, MagicMock)
  assert jobResetAgent.enabled
  assert jobResetAgent.addressFrom == "ilcdirac-admin@cern.ch"
  assert jobResetAgent.userJobTypes == ['User']
  assert jobResetAgent.prodJobTypes == ['MCGeneration', 'MCSimulation', 'MCReconstruction',
                                        'MCReconstruction_Overlay', 'Split', 'MCSimulation_ILD',
                                        'MCReconstruction_ILD', 'MCReconstruction_Overlay_ILD',
                                        'Split_ILD']


def test_begin_execution(jobResetAgent):
  """Test for beginExecution function."""
  jobResetAgent.accounting["Junk"].append("Funk")
  jobResetAgent.am_setOption = MagicMock()
  jobResetAgent.am_getOption = MagicMock()
  getOptionCalls = [call('EnableFlag', False),
                    call('MailTo', jobResetAgent.addressTo),
                    call('MailFrom', jobResetAgent.addressFrom),
                    call('UserJobs', jobResetAgent.userJobTypes),
                    call('ProdJobs', jobResetAgent.prodJobTypes)]

  jobResetAgent.beginExecution()
  jobResetAgent.am_getOption.assert_has_calls(getOptionCalls)
  # accounting dictionary should be cleared
  assert jobResetAgent.accounting == {}


def test_get_jobs(jobResetAgent, today):
  """Test for getJobs function."""
  jobStatus = "Done"
  jobType = "User"
  minorStatus = "Requests Done"
  attrDict = {"JobType": jobType,
              "MinorStatus": minorStatus,
              "Status": jobStatus}

  jobResetAgent.jobDB.selectJobs.return_value = S_ERROR()
  res = jobResetAgent.getJobs(jobStatus, jobType, minorStatus)
  assert not res["OK"]

  jobResetAgent.jobDB.selectJobs.reset_mock()
  jobResetAgent.jobDB.selectJobs.return_value = S_OK(["1", "2", "3"])
  res = jobResetAgent.getJobs(jobStatus, jobType, minorStatus)
  assert res["Value"] == [1, 2, 3]
  jobResetAgent.jobDB.selectJobs.assert_called_once_with(attrDict, older=today - timedelta(days=1))


def test_treat_User_Job_With_No_Req(jobResetAgent, fakeJobID):
  """Test for treatUserJobWithNoReq function."""
  jobResetAgent.markJob = MagicMock()

  # case if getJobsMinorStatus function returns an error
  jobResetAgent.jobMonClient.getJobsMinorStatus.return_value = S_ERROR()
  res = jobResetAgent.treatUserJobWithNoReq(fakeJobID)
  assert not res["OK"]
  jobResetAgent.jobMonClient.getJobsMinorStatus.assert_called_once_with([fakeJobID])

  # case if getJobsMinorStatus executes successfully but getJobsApplicationStatus returns an error
  jobResetAgent.jobMonClient.getJobsMinorStatus.return_value = S_OK({fakeJobID: {'MinorStatus':
                                                                                 JRA.FINAL_MINOR_STATES[0],
                                                                                 'JobID': fakeJobID}})
  jobResetAgent.jobMonClient.getJobsApplicationStatus.return_value = S_ERROR()
  res = jobResetAgent.treatUserJobWithNoReq(fakeJobID)
  assert not res["OK"]

  # mark job done if ApplicationStatus and MinorStatus are in Final States
  jobResetAgent.jobMonClient.getJobsApplicationStatus.return_value = S_OK({fakeJobID:
                                                                           {'ApplicationStatus':
                                                                            JRA.FINAL_APP_STATES[0],
                                                                            'JobID': fakeJobID}})
  res = jobResetAgent.treatUserJobWithNoReq(fakeJobID)
  assert res["OK"]
  jobResetAgent.markJob.assert_called_once_with(fakeJobID, "Done")

  # dont do anything if ApplicationStatus and MinorStatus are not in Final States
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.jobMonClient.getJobsMinorStatus.return_value = S_OK({fakeJobID:
                                                                     {'MinorStatus': 'other status',
                                                                      'JobID': fakeJobID}})
  jobResetAgent.jobMonClient.getJobsApplicationStatus.return_value = S_OK({fakeJobID:
                                                                           {'ApplicationStatus': 'other status',
                                                                            'JobID': fakeJobID}})
  res = jobResetAgent.treatUserJobWithNoReq(fakeJobID)
  assert res["OK"]
  jobResetAgent.markJob.assert_not_called()


def test_treat_User_Job_With_Req(jobResetAgent, fakeJobID):
  """Test for treatUserJobWithReq function."""
  doneRequest = createRequest(requestID=1, opType="RemoveFile", opStatus="Done", fileStatus="Done")
  failedRequestID = 2
  failedRequest = createRequest(requestID=failedRequestID, opType="RemoveFile", opStatus="Failed",
                                fileStatus="Failed")
  jobResetAgent.resetRequest = MagicMock()
  jobResetAgent.markJob = MagicMock()

  # if request status is 'Done' then job should also be marked 'Done'
  jobResetAgent.treatUserJobWithReq(fakeJobID, doneRequest)
  jobResetAgent.markJob.assert_called_once_with(fakeJobID, "Done")
  jobResetAgent.resetRequest.assert_not_called()

  # if request status is not 'Done' then reset request
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.treatUserJobWithReq(fakeJobID, failedRequest)
  jobResetAgent.markJob.assert_not_called()
  jobResetAgent.resetRequest.assert_called_once_with(failedRequestID)

  # if request is waiting
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.resetRequest.reset_mock()
  alreadyRegisteredRequest = createRequest(requestID=4, opType="RegisterFile",
                                           opStatus="Waiting", fileStatus="Waiting")
  assert jobResetAgent.treatUserJobWithReq(fakeJobID, alreadyRegisteredRequest)['OK']
  jobResetAgent.markJob.assert_not_called()
  jobResetAgent.resetRequest.assert_not_called()


def test_treat_User_Job_With_ReplReg(jobResetAgent, fakeJobID):
  # ReplicateAndRegister with No such file
  jobResetAgent.markJob = MagicMock(return_value=S_OK())
  jobResetAgent.resetRequest = MagicMock()
  failedRequest = createRequest(requestID=1, opType='ReplicateAndRegister', opStatus='Failed',
                                fileStatus='Failed', lfnError='No such file or directory', lfn='/ilc/fake/file')

  jobResetAgent.treatUserJobWithReq(fakeJobID, failedRequest)
  jobResetAgent.markJob.assert_called_once_with(fakeJobID, 'Failed')
  jobResetAgent.resetRequest.assert_not_called()

  # ReplicateAndRegister with some other error
  jobResetAgent.markJob = MagicMock()
  jobResetAgent.resetRequest = MagicMock()
  failedRequest = createRequest(requestID=1, opType='ReplicateAndRegister', opStatus='Failed',
                                fileStatus='Failed', lfnError='Failed to replicate', lfn='/ilc/fake/file')

  jobResetAgent.treatUserJobWithReq(fakeJobID, failedRequest)
  jobResetAgent.markJob.assert_not_called()
  jobResetAgent.resetRequest.assert_called_once_with(1)


def test_treat_Failed_Prod_With_Req(jobResetAgent, doneRemoveRequest, doneReplicateRequest,
                                    failedRemoveRequest, failedReplicateRequest):
  """Test for treatFailedProdWithReq function."""
  jobResetAgent.markJob = MagicMock()
  jobResetAgent.resetRequest = MagicMock()
  jobResetAgent.dataManager.removeFile.reset_mock()

  # if request is done then job should be marked failed
  jobResetAgent.treatFailedProdWithReq(fakeJobID, doneRemoveRequest)
  jobResetAgent.markJob.assert_called_once_with(fakeJobID, "Failed")

  jobResetAgent.markJob.reset_mock()
  jobResetAgent.treatFailedProdWithReq(fakeJobID, doneReplicateRequest)
  jobResetAgent.markJob.assert_called_once_with(fakeJobID, "Failed")

  # failed requests with removeFile operation should be reset
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.treatFailedProdWithReq(fakeJobID, failedRemoveRequest)
  fileLfn = failedRemoveRequest[0][0].LFN
  jobResetAgent.dataManager.removeFile.assert_called_once_with([fileLfn], force=True)
  jobResetAgent.resetRequest.assert_called_once_with(getattr(failedRemoveRequest, "RequestID"))
  jobResetAgent.markJob.assert_not_called()

  # failed requests with operations other than removeFile should not be reset
  jobResetAgent.resetRequest.reset_mock()
  jobResetAgent.dataManager.reset_mock()
  jobResetAgent.treatFailedProdWithReq(fakeJobID, failedReplicateRequest)
  jobResetAgent.dataManager.assert_not_called()
  jobResetAgent.resetRequest.assert_not_called()
  jobResetAgent.markJob.assert_not_called()


def test_treat_Failed_Prod_With_No_Req(jobResetAgent, fakeJobID):
  """Test for treatFailedProdWithNoReq function."""
  jobResetAgent.markJob = MagicMock()
  jobResetAgent.treatFailedProdWithNoReq(fakeJobID)
  jobResetAgent.markJob.assert_called_once_with(fakeJobID, "Failed")


def test_treat_Completed_Prod_With_Req(jobResetAgent, fakeJobID, doneRemoveRequest, doneReplicateRequest,
                                       failedRemoveRequest, failedReplicateRequest, failedRemoveReplicaRequest):
  """Test for treatCompletedProdWithReq function."""
  jobResetAgent.markJob = MagicMock()
  jobResetAgent.resetRequest = MagicMock()

  # if request is done then job should be marked Done
  jobResetAgent.treatCompletedProdWithReq(fakeJobID, doneRemoveRequest)
  jobResetAgent.markJob.assert_called_once_with(fakeJobID, "Done")

  jobResetAgent.markJob.reset_mock()
  jobResetAgent.treatCompletedProdWithReq(fakeJobID, doneReplicateRequest)
  jobResetAgent.markJob.assert_called_once_with(fakeJobID, "Done")

  # job with failed ReplicateAndRegister operation should be marked done if file does not exist
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.markJob = MagicMock(return_value=S_OK())
  request = createRequest(requestID=1, opType="RemoveFile", opStatus="Done",
                          fileStatus="Done", lfnError="No such file")
  jobResetAgent.treatCompletedProdWithReq(fakeJobID, request)
  jobResetAgent.markJob.assert_called_once_with(fakeJobID, "Done")

  # failed requests with ReplicateAndRegister operation should be reset
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.treatCompletedProdWithReq(fakeJobID, failedReplicateRequest)
  jobResetAgent.resetRequest.assert_called_once_with(getattr(failedReplicateRequest, "RequestID"))

  # failed Remove file request should not be reset
  jobResetAgent.resetRequest.reset_mock()
  jobResetAgent.treatCompletedProdWithReq(fakeJobID, failedRemoveRequest)
  jobResetAgent.markJob.assert_not_called()
  jobResetAgent.resetRequest.assert_not_called()

  # failed Remove file request should not be reset
  jobResetAgent.resetRequest.reset_mock()
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.treatCompletedProdWithReq(fakeJobID, failedRemoveReplicaRequest)
  jobResetAgent.markJob.assert_called_once_with(fakeJobID, 'Done')
  jobResetAgent.resetRequest.assert_not_called()

  # is waiting
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.resetRequest.reset_mock()
  alreadyRegisteredRequest = createRequest(requestID=4, opType="RegisterFile",
                                           opStatus="Waiting", fileStatus="Waiting")
  assert jobResetAgent.treatCompletedProdWithReq(fakeJobID, alreadyRegisteredRequest)['OK']
  jobResetAgent.markJob.assert_not_called()
  jobResetAgent.resetRequest.assert_not_called()

  # request failing because no such file, mark it done
  jobResetAgent.accounting.clear()
  request = createRequest(requestID=1, opType="ReplicateAndRegister", opStatus="Failed",
                          fileStatus="Failed", lfnError="No such file", lfn="/ilc/fake/file")
  jobResetAgent.markJob = MagicMock(return_value=S_OK())
  jobResetAgent.treatCompletedProdWithReq(1234, request)
  jobResetAgent.markJob.assert_called_once_with(1234, "Done")
  assert len(jobResetAgent.accounting["Production"]) == 1
  assert jobResetAgent.accounting["Production"][0]["JobID"] == 1234


def test_treat_Completed_Prod_With_No_Req(jobResetAgent, fakeJobID):
  """Test for treatCompletedProdWithNoReq function."""
  jobResetAgent.markJob = MagicMock()
  jobResetAgent.treatCompletedProdWithNoReq(fakeJobID)
  jobResetAgent.markJob.assert_called_once_with(fakeJobID, "Done")


def test_check_jobs(jobResetAgent):
  """Test for checkJobs function."""
  jobIDs = [1, 2]
  dummy_treatJobWithNoReq = MagicMock()
  dummy_treatJobWithReq = MagicMock()

  # if the readRequestsForJobs func returns error than checkJobs should exit and return an error
  jobResetAgent.reqClient.readRequestsForJobs.return_value = S_ERROR()
  res = jobResetAgent.checkJobs(jobIDs, treatJobWithNoReq=dummy_treatJobWithNoReq,
                                treatJobWithReq=dummy_treatJobWithReq)
  assert not res["OK"]

  # test if correct treatment functions are called
  jobResetAgent.reqClient.readRequestsForJobs.return_value = S_OK({'Successful': {},
                                                                   'Failed': {jobIDs[0]: 'Request not found'}})
  jobResetAgent.checkJobs(jobIDs, treatJobWithNoReq=dummy_treatJobWithNoReq,
                          treatJobWithReq=dummy_treatJobWithReq)
  dummy_treatJobWithNoReq.assert_has_calls([call(jobIDs[0]), call(jobIDs[1])])
  dummy_treatJobWithReq.assert_not_called()

  dummy_treatJobWithNoReq.reset_mock()
  req1 = Request({"RequestID": 1})
  req2 = Request({"RequestID": 2})
  jobResetAgent.reqClient.readRequestsForJobs.return_value = S_OK({'Successful': {jobIDs[0]: req1,
                                                                                  jobIDs[1]: req2},
                                                                   'Failed': {}})
  jobResetAgent.checkJobs(jobIDs, treatJobWithNoReq=dummy_treatJobWithNoReq,
                          treatJobWithReq=dummy_treatJobWithReq)
  dummy_treatJobWithNoReq.assert_not_called()
  dummy_treatJobWithReq.assert_has_calls([call(jobIDs[0], req1), call(jobIDs[1], req2)])


def test_get_staged_files(jobResetAgent):
  """Test for getStagedFiles function."""
  stagedFile = "/ilc/fake/lfn1/staged"
  nonStagedFile = "/ilc/fake/lfn2/nonStaged"
  lfns = [stagedFile, nonStagedFile]

  res = jobResetAgent.getStagedFiles([])
  assert res["OK"]

  SeModule.StorageElementItem.getFileMetadata = MagicMock(return_value=S_ERROR())
  res = jobResetAgent.getStagedFiles(lfns)
  assert not res["OK"]

  SeModule.StorageElementItem.getFileMetadata.return_value = S_OK({'Successful': {stagedFile: {'Cached': 1},
                                                                                  nonStagedFile: {'Cached': 0}}})
  res = jobResetAgent.getStagedFiles(lfns)
  assert res["Value"] == [stagedFile]


def test_get_input_data_for_jobs(jobResetAgent):
  """Test for getInputDataForJobs function."""
  jobIDs = [1, 2]
  lfn1 = "lfn:/ilc/fake/lfn1"
  lfn2 = "/ilc/fake/lfn2"
  jobResetAgent.jobMonClient.getInputData.return_value = S_ERROR()

  res = jobResetAgent.getInputDataForJobs(jobIDs)
  assert res["Value"] == {}

  jobResetAgent.jobMonClient.getInputData.return_value = S_OK([lfn1, lfn2])
  res = jobResetAgent.getInputDataForJobs(jobIDs)
  assert res["Value"] == {lfn1[4:]: jobIDs, lfn2: jobIDs}


def test_reschedule_jobs(jobResetAgent):
  """Test for rescheduleJobs function."""
  jobShouldFailToReset = 1
  jobShouldSuccessfullyReset = 2
  jobsToReschedule = [jobShouldFailToReset, jobShouldSuccessfullyReset]

  jobResetAgent.jobManagerClient.resetJob.side_effect = [S_ERROR(), S_OK()]
  res = jobResetAgent.rescheduleJobs(jobsToReschedule)
  assert res['OK']
  assert res["Value"]["Successful"] == [jobShouldSuccessfullyReset]
  assert res["Value"]["Failed"] == [jobShouldFailToReset]


def test_check_staging_jobs(jobResetAgent):
  """Test for checkStagingJobs function."""
  jobShouldBeRescheduled = 1
  jobShouldNotBeResecheduled = 2
  stagedFile = "/ilc/file/staged"
  notStagedFile = "/ilc/file/notStaged"
  jobIDs = [jobShouldBeRescheduled, jobShouldNotBeResecheduled]

  jobResetAgent.getInputDataForJobs = MagicMock()
  jobResetAgent.getStagedFiles = MagicMock()
  jobResetAgent.rescheduleJobs = MagicMock()

  jobResetAgent.getInputDataForJobs.return_value = S_OK({})
  res = jobResetAgent.checkStagingJobs(jobIDs)
  assert res['OK']
  jobResetAgent.getInputDataForJobs.assert_called_once_with(jobIDs)
  jobResetAgent.getStagedFiles.assert_not_called()

  jobsToReschedule = set()
  jobsToReschedule.add(jobShouldBeRescheduled)
  jobResetAgent.getInputDataForJobs.reset_mock()
  jobResetAgent.getInputDataForJobs.return_value = S_OK({stagedFile: jobShouldBeRescheduled,
                                                         notStagedFile: jobShouldNotBeResecheduled})
  jobResetAgent.getStagedFiles.return_value = S_OK([stagedFile])
  jobResetAgent.checkStagingJobs(jobIDs)
  jobResetAgent.rescheduleJobs.assert_called_once_with(jobsToReschedule)


def test_reset_request(jobResetAgent):
  """Test for resetRequest function."""
  fakeReqID = 1
  jobResetAgent.logError = MagicMock()
  jobResetAgent.reqClient.resetFailedRequest.return_value = S_ERROR()
  res = jobResetAgent.resetRequest(fakeReqID)
  jobResetAgent.logError.assert_called()
  assert not res["OK"]

  jobResetAgent.logError.reset_mock()
  jobResetAgent.reqClient.resetFailedRequest.return_value = S_OK("Not reset")
  res = jobResetAgent.resetRequest(fakeReqID)
  assert not res["OK"]
  jobResetAgent.logError.assert_called()

  jobResetAgent.logError.reset_mock()
  jobResetAgent.reqClient.resetFailedRequest.return_value = S_OK()
  res = jobResetAgent.resetRequest(fakeReqID)
  assert res['OK']
  jobResetAgent.logError.assert_not_called()


def test_mark_job(jobResetAgent):
  """Test for markJob function."""
  fakeMinorStatus = "fakeMinorStatus"
  fakeApp = "fakeApp"
  fakeJobStatus = "Done"
  defaultMinorStatus = "Requests Done"
  defaultApplication = "CompletedJobChecker"

  # default minorStatus should be "Requests Done" and application should be "CompletedJobChecker"
  jobResetAgent.jobStateUpdateClient.setJobStatus = MagicMock(return_value=S_ERROR())
  res = jobResetAgent.markJob(fakeJobID, fakeJobStatus)
  assert not res["OK"]
  jobResetAgent.jobStateUpdateClient.setJobStatus.assert_called_once_with(fakeJobID,
                                                                          fakeJobStatus,
                                                                          defaultMinorStatus,
                                                                          defaultApplication)

  jobResetAgent.jobStateUpdateClient.setJobStatus.reset_mock()
  jobResetAgent.jobStateUpdateClient.setJobStatus.return_value = S_OK()
  res = jobResetAgent.markJob(fakeJobID, fakeJobStatus, minorStatus=fakeMinorStatus, application=fakeApp)
  assert res['OK']
  jobResetAgent.jobStateUpdateClient.setJobStatus.assert_called_once_with(fakeJobID, fakeJobStatus,
                                                                          fakeMinorStatus, fakeApp)


def test_execute(jobResetAgent):
  """Test for execute function."""
  jobIDs = [1, 2]
  jobResetAgent.getJobs = MagicMock()
  jobResetAgent.checkJobs = MagicMock()
  jobResetAgent.checkStagingJobs = MagicMock()

  jobResetAgent.getJobs.return_value = S_OK(jobIDs)
  jobResetAgent.execute()
  # check if checkJobs function is called with correct arguments
  completedProdJobCall = call(jobIDs=jobIDs, treatJobWithNoReq=jobResetAgent.treatCompletedProdWithNoReq,
                              treatJobWithReq=jobResetAgent.treatCompletedProdWithReq)
  failedProdJobCall = call(jobIDs=jobIDs, treatJobWithNoReq=jobResetAgent.treatFailedProdWithNoReq,
                           treatJobWithReq=jobResetAgent.treatFailedProdWithReq)
  completedUserJob = call(jobIDs=jobIDs, treatJobWithNoReq=jobResetAgent.treatUserJobWithNoReq,
                          treatJobWithReq=jobResetAgent.treatUserJobWithReq)
  calls = [completedProdJobCall, failedProdJobCall, completedUserJob]
  jobResetAgent.checkJobs.assert_has_calls(calls)
  jobResetAgent.checkStagingJobs.assert_called_once_with(jobIDs)


def test_send_notification(jobResetAgent):
  """Test for sendNotification function."""
  jobResetAgent.errors = []
  jobResetAgent.accounting = {}

  # send mail should not be called if there are no errors and accounting information
  jobResetAgent.sendNotification()
  jobResetAgent.nClient.sendMail.assert_not_called()

  # send mail should be called if there are errors but no accounting information
  jobResetAgent.errors = ["some error"]
  jobResetAgent.sendNotification()
  jobResetAgent.nClient.sendMail.assert_called()

  # send email should be called if there is accounting information but no errors
  jobResetAgent.nClient.sendMail.reset_mock()
  jobResetAgent.errors = []
  jobResetAgent.accounting = {"User": [{"JobID": 123, "JobStatus": "Failed", "Treatment": "reset request"}],
                              "Prod": [{"JobID": 124, "JobStatus": "Failed", "Treatment": "reset request"}]}
  jobResetAgent.sendNotification()
  jobResetAgent.nClient.sendMail.assert_called()

  # try sending email to all addresses even if we get error for sending email to some address
  jobResetAgent.nClient.sendMail.reset_mock()
  jobResetAgent.errors = ["some error"]
  jobResetAgent.addressTo = ["name1@cern.ch", "name2@cern.ch"]
  jobResetAgent.nClient.sendMail.return_value = S_ERROR()
  jobResetAgent.sendNotification()
  assert len(jobResetAgent.nClient.sendMail.mock_calls) == len(jobResetAgent.addressTo)

  # accounting dict and errors list should be cleared after notification is sent
  assert jobResetAgent.accounting == {}
  assert jobResetAgent.errors == []


def test_treat_User_Job_With_RegisterFile(jobResetAgent, userProxyFixture, fakeJobID, alreadyErrorMessage):
  """Test user jobs with RegisterFile request operations."""
  jobResetAgent.markJob = MagicMock()
  jobResetAgent.resetRequest = MagicMock()

  # if Already registered file request
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.resetRequest.reset_mock()
  jobResetAgent._fcClient.removeFile.return_value = S_OK(dict(Failed={},
                                                              Successful={'/ilc/fake/lfn': True}))
  alreadyRegisteredRequest = createRequest(requestID=4, opType="RegisterFile",
                                           opStatus="Failed", fileStatus="Failed",
                                           lfnError=alreadyErrorMessage)
  jobResetAgent.treatUserJobWithReq(fakeJobID, alreadyRegisteredRequest)
  jobResetAgent.markJob.assert_not_called()
  jobResetAgent.resetRequest.assert_called_once_with(4)
  jobResetAgent._fcClient.removeFile.assert_called_once_with('/ilc/fake/lfn')

  # request is waiting
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.resetRequest.reset_mock()
  jobResetAgent._fcClient.removeFile.reset_mock()
  alreadyRegisteredRequest = createRequest(requestID=4, opType="RegisterFile",
                                           opStatus="Waiting", fileStatus="Waiting",
                                           lfnError=None)
  jobResetAgent.treatUserJobWithReq(fakeJobID, alreadyRegisteredRequest)
  jobResetAgent.markJob.assert_not_called()
  jobResetAgent.resetRequest.assert_not_called()
  jobResetAgent._fcClient.removeFile.assert_not_called()

  # request is waiting
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.resetRequest.reset_mock()
  jobResetAgent._fcClient.removeFile.reset_mock()
  alreadyRegisteredRequest = createRequest(requestID=4, opType="RegisterFile",
                                           opStatus="Failed", fileStatus="Failed",
                                           lfnError="File not found")
  jobResetAgent.treatUserJobWithReq(fakeJobID, alreadyRegisteredRequest)
  jobResetAgent.markJob.assert_not_called()
  jobResetAgent._fcClient.removeFile.assert_not_called()

  # removeFile failed
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.resetRequest.reset_mock()
  jobResetAgent._fcClient.removeFile.reset_mock()
  jobResetAgent._fcClient.removeFile.return_value = S_OK(dict(Failed={'/ilc/fake/lfn': "not author"},
                                                              Successful={}))
  alreadyRegisteredRequest = createRequest(requestID=4, opType="RegisterFile",
                                           opStatus="Failed", fileStatus="Failed",
                                           lfnError=alreadyErrorMessage)
  res = jobResetAgent.treatUserJobWithReq(fakeJobID, alreadyRegisteredRequest)
  assert not res['OK']
  jobResetAgent.markJob.assert_not_called()
  jobResetAgent.resetRequest.assert_not_called()
  jobResetAgent._fcClient.removeFile.assert_called_once_with("/ilc/fake/lfn")

  # removeFile failed with S_ERROR
  jobResetAgent.markJob.reset_mock()
  jobResetAgent.resetRequest.reset_mock()
  jobResetAgent._fcClient.removeFile.reset_mock()
  jobResetAgent._fcClient.removeFile.return_value = S_ERROR("Failed")
  alreadyRegisteredRequest = createRequest(requestID=4, opType="RegisterFile",
                                           opStatus="Failed", fileStatus="Failed",
                                           lfnError=alreadyErrorMessage)
  res = jobResetAgent.treatUserJobWithReq(fakeJobID, alreadyRegisteredRequest)
  assert not res['OK']
  jobResetAgent.markJob.assert_not_called()
  jobResetAgent.resetRequest.assert_not_called()
  jobResetAgent._fcClient.removeFile.assert_called_once_with("/ilc/fake/lfn")


def test_treat_User_Job_With_RegisterFile_2(jobResetAgent, failingUserProxyFixture, fakeJobID, alreadyErrorMessage):
  """Test when UserProxy failed."""
  jobResetAgent.markJob = MagicMock()
  jobResetAgent.resetRequest = MagicMock()

  jobResetAgent.markJob.reset_mock()
  jobResetAgent.resetRequest.reset_mock()
  jobResetAgent._fcClient.removeFile.reset_mock()
  jobResetAgent._fcClient.removeFile.return_value = S_ERROR("Failed")
  alreadyRegisteredRequest = createRequest(requestID=4, opType="RegisterFile",
                                           opStatus="Failed", fileStatus="Failed",
                                           lfnError=alreadyErrorMessage)
  res = jobResetAgent.treatUserJobWithReq(fakeJobID, alreadyRegisteredRequest)
  assert not res['OK']
  jobResetAgent.markJob.assert_not_called()
  jobResetAgent.resetRequest.assert_not_called()
  jobResetAgent._fcClient.removeFile.assert_not_called()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""CalibrationAgent.

Supervises the state of the jobs started by the CalibrationService and requests resubmission if
too many failed.
"""
from __future__ import absolute_import
from collections import defaultdict

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Base.Client import Client
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient
from DIRAC.WorkloadManagementSystem.Client.WMSClient import WMSClient
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from ILCDIRAC.CalibrationSystem.Utilities.functions import convert_to_int_list

__RCSID__ = "$Id$"


class CalibrationAgent(AgentModule):
  """CalibrationAgent.

  Periodically checks the state of the jobs from the CalibrationService.
  If too few jobs are running, it tells the CalibrationService to restart the failed ones.
  """

  def initialize(self):
    """Initialize the Agent."""
    self.calibrationService = Client()
    self.calibrationService.setServer('Calibration/Calibration')
    self.currentCalibrations = []  # Contains IDs (int) of the calibrations
    self.currentJobStatusesPerWorker = {}  # Contains a mapping calibrationID -> dict, the dict contains a mapping
    self.currentJobStatusesPerJobId = {}
    self.calibrationOwnership = {}
    self.ops = Operations()
    # WorkerID (int) -> jobStatus (enum)
    return S_OK()

  def execute(self):
    """Execute one per cycle.

    Minimum time between two cycles is set via the PollingTime
    parameter (in seconds) in the ConfigTemplate.cfg
    """
    # TODO: Implement: call fetchJobStatuses, compare this to old status. If an entire calibrationRun finished,
    # remove it from data structures. If too many jobs failed, ask Service for resubmission. Then replace old
    # job status dict with new one
    # To clear up: Can a job disappear from this list? Or what happens if node crashes.
    res = self.fetchJobStatuses()
    if not res['OK']:
      return res
    self.currentJobStatusesPerWorker = res['Value']['jobStatusVsWorkerId']
    self.currentJobStatusesPerJobId = res['Value']['jobStatusVsJobId']
    self.calibrationOwnership = res['Value']['calibrationOwnership']
    self.checkForCalibrationsToBeKilled()
    res = self.calibrationService.getNumberOfJobsPerCalibration()
    if not res['OK']:
      return res
    targetJobNumbers = res['Value']
    # FIXME self.currentCalibrations is not used anywhere...
    self.currentCalibrations = list(targetJobNumbers.keys())

    res = self.calibrationService.getRunningCalibrations()
    if not res['OK']:
      return res
    runningCalibs = res['Value']  # list of unfinished calibrations
    currentJobStatusesPerWorker_runningCalibs = {}
    targetJobNumbers_runningCalibs = {}
    for iCalib in runningCalibs:
      try:
        currentJobStatusesPerWorker_runningCalibs[iCalib] = self.currentJobStatusesPerWorker[iCalib]
        targetJobNumbers_runningCalibs[iCalib] = targetJobNumbers[iCalib]
      except KeyError:
        errMsgConst = 'Error while retrieving information for calibration'
        errMsgVariable = '#%s' % iCalib
        self.log.error(errMsgConst, errMsgVariable)

    jobsToResubmitted = CalibrationAgent.__calculateJobsToBeResubmitted(
        currentJobStatusesPerWorker_runningCalibs, targetJobNumbers_runningCalibs)

    if jobsToResubmitted:
      res = self.requestResubmission(jobsToResubmitted)
      if not res['OK']:
        return res

    res = self.calibrationService.checkForStepIncrement()
    if not res['OK']:
      return res
    return S_OK()

  def fetchJobStatuses(self):
    """Request the statuses of jobs of all active calibrations.

    Output is mapped as calibrationID -> workerID -> jobStatus.

    :returns: Dictionary of type calibrationID -> dict, with dict of type workerID (int) -> jobStatus (enum)
    :rtype: dict
    """
    jobMonitoringService = JobMonitoringClient()
    jobIDs = []
    res = self.calibrationService.getActiveCalibrations()
    if not res['OK']:
      return res
    activeCalibrations = res['Value']
    self.log.debug("jobMonitoringService:", " %s" % jobMonitoringService)
    self.log.debug("self.calibrationService:", " %s" % self.calibrationService)
    self.log.debug("activeCalibrations:", " %s" % activeCalibrations)
    for iCalib in activeCalibrations:
      res = jobMonitoringService.getJobs({'JobGroup': 'PandoraCaloCalibration_calid_%s' % iCalib})
      if not res['OK']:
        self.log.error("Failed getting job IDs from job DB! Error:", res['Message'])
        return S_ERROR('Failed getting job IDs from job DB!')
      jobIDs += res['Value']
    self.log.debug('jobIDs:', ' %s' % jobIDs)
    res = jobMonitoringService.getJobsParameters(convert_to_int_list(
        jobIDs), ['JobName', 'Status', 'JobID', 'Owner', 'OwnerGroup', 'OwnerDN'])
    if not res['OK']:
      self.log.error("Cannot retrieve jobs parameters from job monitoring service", res['Message'])
      return res
    jobStatuses = res['Value']
    result1 = defaultdict(dict)  # defaults to {}
    result2 = defaultdict(dict)  # defaults to {}
    result3 = defaultdict(dict)  # defaults to {}
    for attrDict in jobStatuses.values():
      jobName = attrDict['JobName']
      curCalibration = CalibrationAgent.__getCalibrationIDFromJobName(jobName)
      result1[curCalibration].update({CalibrationAgent.__getWorkerIDFromJobName(jobName):
                                      attrDict['Status']})
      result2[curCalibration].update({attrDict['JobID']: attrDict['Status']})
      if curCalibration not in result3:
        result3[curCalibration].update({'Owner': attrDict['Owner']})
        result3[curCalibration].update({'OwnerGroup': attrDict['OwnerGroup']})
        result3[curCalibration].update({'OwnerDN': attrDict['OwnerDN']})
    return S_OK({'jobStatusVsWorkerId': dict(result1), 'jobStatusVsJobId': dict(result2),
                 'calibrationOwnership': dict(result3)})

  def sendKillSignalToJobManager(self, jobIdsToKill, iCalibId):
    """Send kill signals to job manager service."""
    jobManagerService = WMSClient(useCertificates=True, delegatedDN=self.calibrationOwnership[iCalibId]['OwnerDN'],
                                  delegatedGroup=self.calibrationOwnership[iCalibId]['OwnerGroup'])
    res = jobManagerService.killJob(jobIdsToKill)
    return res

  def checkForCalibrationsToBeKilled(self):
    """Ask calibation service for jobs to be killed and kill them."""
    res = self.calibrationService.getCalibrationsToBeKilled()
    if not res['OK']:
      self.log.error('Failed to get list of calibrations to be killed from service. errMsg:', '%s' % res['Message'])
      return S_OK()
    calibIds = res['Value']
    if len(calibIds) == 0:
      return S_OK()
    else:
      for iCalibId in calibIds:
        if iCalibId not in self.currentJobStatusesPerJobId:
          self.log.info('No jobs to kill for calibration', '#%s' % iCalibId)
        else:
          jobIdsToKill = list(self.currentJobStatusesPerJobId[iCalibId].keys())
          self.sendKillSignalToJobManager(jobIdsToKill, iCalibId)
          if not res['OK']:
            self.log.error('Failed to kill jobs. errMsg:', '%s' % res['Message'])
          else:
            self.log.info('Kill jobs which belong to calibrations:', '%s' % res['Value'])
    return S_OK()

  def requestResubmission(self, failedJobs):
    """Request the Service to resubmit the failed jobs.

    :param list failedJobs: List of 2-tuples ( calibrationID, workerID )
    :returns: None
    """
    jobs_to_resubmit = failedJobs
    number_of_tries = 0
    result = S_ERROR()
    self.resubmissionRetries = self.ops.getValue('Calibration/ResubmissionRetries', 5)  # TODO add this to CS
    while not result['OK'] and number_of_tries < self.resubmissionRetries:
      result = self.calibrationService.resubmitJobs(jobs_to_resubmit)
      number_of_tries += 1
      # FIXME: ResubmitJobs will probably be implemented in a way that would allow some resubmissions to fail and some
      # to work. Thus, this method would need a list of all resubmissions that have yet to be done, which is updated
      # in each iteration. once it is empty, the method returns. If it takes too long, RuntimeError is raised
      if result['OK']:
        return S_OK()
      #  else:
      #    jobs_to_resubmit = result['failed_pairs']
    raise RuntimeError('Cannot resubmit the necessary failed jobs. Problem: %s' % result)

  @staticmethod
  def __calculateJobsToBeResubmitted(jobStatusDict, targetNumberDict):
    """Return list of worker node ids in each calibration where job execution failed.

    :param dict jobStatusDict: Dictionary with a mapping from calibrationID -> dict, with dict having a mapping
                               workerID -> jobStatus
    :param dict targetNumberDict: Dictionary with a mapping from calibrationID -> number of jobs originally alotted
                                  to the calibration
    :returns: List containing 2-tuples ( calibrationID, workerID )
    :rtype: list
    """
    JOB_STATUS_FAILED = ['Failed', 'Killed', 'Stalled']
    failedJobs = []
    for calibrationID, workerDict in jobStatusDict.items():
      for workerID, jobStatus in workerDict.items():
        if jobStatus in JOB_STATUS_FAILED:
          failedJobs.append((calibrationID, workerID))
    return failedJobs

  @staticmethod
  def __getWorkerIDFromJobName(jobname):
    """Extract the worker ID from the raw job name.

    :param str jobname: name of the job in the DIRAC DB
    :returns: the worker ID contained in the name string
    :rtype: int
    """
    return int(jobname.split('_')[4])

  @staticmethod
  def __getCalibrationIDFromJobName(jobname):
    """Extract the calibration ID from the raw job name.

    :param str jobname: name of the job in the DIRAC DB
    :returns: the calibration ID contained in the name string
    :rtype: int
    """
    return int(jobname.split('_')[2])



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Define a new WHIZARD version.

Update the process list, create the new tar ball, upload it, upload the new processlist.

:since: Sep 21, 2010

:author: sposs, sailer, strube
"""

from __future__ import print_function
from __future__ import absolute_import
from six.moves import input
__RCSID__ = "$Id$"

from DIRAC import gConfig, gLogger, exit as dexit
from DIRAC.Core.Base import Script
from DIRAC.Core.Utilities.DIRACScript import DIRACScript
from DIRAC import S_OK, S_ERROR
import os
import tarfile
import shutil
import sys
import subprocess

try:
  import hashlib as md5
except ImportError:
  import md5


class _Params(object):
  def __init__(self):
    self.path = ''
    self.version = ''
    self.platform = 'x86_64-slc5-gcc43-opt'
    self.beam_spectra = ''

  def setVersion(self, optionValue):
    self.version = optionValue
    return S_OK()

  def setPlatform(self, optionValue):
    self.platform = optionValue
    return S_OK()

  def setBeamSpectra(self, optionValue):
    self.beam_spectra = optionValue
    return S_OK()

  def setPath(self, optionValue):
    self.path = optionValue
    return S_OK()

  def registerSwitches(self):
    Script.registerSwitch('P:', "Platform=", 'Platform to use', self.setPlatform)
    Script.registerSwitch('p:', "Path=", "Path to the Whizard results directory", self.setPath)
    Script.registerSwitch("V:", "Version=", "Whizard version", self.setVersion)
    Script.registerSwitch('b:', 'BeamSpectra=', 'Beam spectra version', self.setBeamSpectra)
    Script.setUsageMessage('\n'.join([__doc__.split('\n')[1],
                                         '\nUsage:',
                                         '  %s [option|cfgfile] ...\n' % Script.scriptName]))


def redirectLogOutput(dummy_fd, message):
  """Needed to catch the log output of the shellCall below."""
  sys.stdout.flush()
  print(message)


def readPRCFile(prc):
  """Read the prc file to create the process description."""
  myPRCList = {}
  with open(prc) as myprc:
    model = ""
    for process in myprc:
      process = process.rstrip()
      if not len(process):
        continue
      if process[0] == "#":
        continue
      elems = process.split()
      if elems[0] == "alias":
        continue
      elif elems[0] == "model":
        model = elems[1]
      elif not elems[0] == "model":
        myPRCList[elems[0]] = {}
        myPRCList[elems[0]]['Detail'] = "->".join(elems[1:3])
        myPRCList[elems[0]]['Generator'] = elems[3]
        myPRCList[elems[0]]['Restrictions'] = "none"
        if len(elems) > 4:
          myPRCList[elems[0]]['Restrictions'] = " ".join(elems[4:])
        myPRCList[elems[0]]['Model'] = model
        myPRCList[elems[0]]['InFile'] = "whizard.template.in"
      else:
        continue
    return myPRCList


def getDetailsFromPRC(prc, processin):
  """Get the process details from the prc file."""
  details = {}
  with open(prc) as myprc:
    model = ""
    for process in myprc:
      process = process.rstrip()
      if not len(process):
        continue
      elems = process.split()
      if process[0] == "#":
        continue
      elif elems[0] == "model":
        model = elems[1]
      elif not elems[0] == "model":
        if elems[0] == processin:
          details['Model'] = model
          details['Generator'] = elems[3]
          details['Restrictions'] = "none"
          if len(elems) > 4:
            details['Restrictions'] = " ".join(elems[4:])
          break
  return details


def checkGFortranVersion():
  """Find the version of the gfortran compiler."""
  proc = subprocess.Popen(['gfortran', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  out, dummy_err = proc.communicate()
  if out.find("4.4") > -1:
    return S_OK()
  return S_ERROR("Wrong gFortran version")


def checkSLCVersion():
  """Check the version of the operating system compiler."""
  proc = subprocess.Popen(['cat', '/etc/issue'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  out, dummy_err = proc.communicate()
  if out.find("release 5") > -1:
    return S_OK()
  return S_ERROR("Wrong Operating System")


def getListOfLibraries(pathName):
  """Get List of Libraries for library or executable."""
  listOfLibraries = []
  proc = subprocess.Popen(['ldd', pathName], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  out, dummyerr = proc.communicate()
  for line in out.split("\n"):
    fields = line.split("=>")
    if len(fields) == 2:
      newFields = fields[1].split()
      if len(newFields) == 2:
        lib = newFields[0].strip()
        listOfLibraries.append(lib)
  return listOfLibraries


@DIRACScript()
def main():
  """Do the instalation for new whizard version Copy libraries, create tarball, upload processList file add entry in configuration system."""
  res = checkSLCVersion()
  if not res['OK']:
    gLogger.error(res['Message'])
    dexit(1)

  res = checkGFortranVersion()
  if not res['OK']:
    gLogger.error(res['Message'])
    dexit(1)

  cliParams = _Params()
  cliParams.registerSwitches()
  Script.parseCommandLine(ignoreErrors=False)

  whizardResultFolder = cliParams.path
  platform = cliParams.platform
  whizard_version = cliParams.version
  appVersion = whizard_version
  beam_spectra_version = cliParams.beam_spectra

  if not whizardResultFolder or not whizard_version or not beam_spectra_version:
    Script.showHelp()
    dexit(2)

  from ILCDIRAC.Core.Utilities.ProcessList import ProcessList
  from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
  from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin
  from ILCDIRAC.Core.Utilities.FileUtils import upload
  from DIRAC.DataManagementSystem.Client.DataManager import DataManager
  diracAdmin = DiracAdmin()

  modifiedCS = False

  softwareSection = "/Operations/Defaults/AvailableTarBalls"
  processlistLocation = "ProcessList/Location"

  appName = "whizard"

  ops = Operations()
  path_to_process_list = ops.getValue(processlistLocation, "")
  if not path_to_process_list:
    gLogger.error("Could not find process list location in CS")
    dexit(2)

  gLogger.verbose("Getting process list from file catalog")
  datMan = DataManager()
  res = datMan.getFile(path_to_process_list)
  if not res['OK']:
    gLogger.error("Error while getting process list from storage")
    dexit(2)
  gLogger.verbose("done")

  # just the name of the local file in current working directory
  processlist = os.path.basename(path_to_process_list)
  if not os.path.exists(processlist):
    gLogger.error("Process list does not exist locally")
    dexit(2)

  pl = ProcessList(processlist)

  startDir = os.getcwd()
  inputlist = {}
  os.chdir(whizardResultFolder)
  folderlist = os.listdir(whizardResultFolder)

  whiz_here = folderlist.count("whizard")
  if whiz_here == 0:
    gLogger.error("whizard executable not found in %s, please check" % whizardResultFolder)
    os.chdir(startDir)
    dexit(2)

  whizprc_here = folderlist.count("whizard.prc")
  if whizprc_here == 0:
    gLogger.error("whizard.prc not found in %s, please check" % whizardResultFolder)
    os.chdir(startDir)
    dexit(2)

  whizmdl_here = folderlist.count("whizard.mdl")
  if whizmdl_here == 0:
    gLogger.error("whizard.mdl not found in %s, please check" % whizardResultFolder)
    os.chdir(startDir)
    dexit(2)

  gLogger.verbose("Preparing process list")

  # FIXME:: What is this doing exactly? Is this necessary? -- APS, JFS
  for f in folderlist:
    if f.count(".in"):
      infile = open(f, "r")
      found_detail = False

      for line in infile:
        if line.count("decay_description"):
          currprocess = f.split(".template.in")[0]
          inputlist[currprocess] = {}
          inputlist[currprocess]["InFile"] = f.rstrip("~")
          inputlist[currprocess]["Detail"] = line.split("\"")[1]
          found_detail = True
        if line.count("process_id") and found_detail:
          process_id = line.split("\"")[1]
          inputlist[currprocess]["Model"] = ""
          inputlist[currprocess]["Generator"] = ""
          inputlist[currprocess]["Restrictions"] = ""
          for process in process_id.split():
            print("Looking for detail of process %s" % (process))
            process_detail = getDetailsFromPRC("whizard.prc", process)
            inputlist[currprocess]["Model"] = process_detail["Model"]
            inputlist[currprocess]["Generator"] = process_detail["Generator"]
            if len(inputlist[currprocess]["Restrictions"]):
              inputlist[currprocess]["Restrictions"] = inputlist[currprocess]["Restrictions"] + \
                  ", " + process_detail["Restrictions"]
            else:
              inputlist[currprocess]["Restrictions"] = process_detail["Restrictions"]
      # if len(inputlist[currprocess].items()):
      #  inputlist.append(processdict)
  # END FIXEME

  # Update inputlist with what was found looking in the prc file
  processes = readPRCFile("whizard.prc")
  inputlist.update(processes)

  # get from cross section files the cross sections for the processes in inputlist
  # Need full process list
  for f in folderlist:
    if f.count("cross_sections_"):
      crossfile = open(f, "r")
      for line in crossfile:
        line = line.rstrip().lstrip()
        if not len(line):
          continue
        if line[0] == "#" or line[0] == "!":
          continue
        if len(line.split()) < 2:
          continue
        currprocess = line.split()[0]
        if currprocess in inputlist:
          inputlist[currprocess]['CrossSection'] = line.split()[1]

  gLogger.notice("Preparing Tarball")

  # Make a folder in the current directory of the user to store the whizard libraries, executable et al.
  localWhizardFolderRel = ("whizard" + whizard_version)  # relative path
  localWhizardFolder = os.path.join(startDir, localWhizardFolderRel)

  if not os.path.exists(localWhizardFolder):
    os.makedirs(localWhizardFolder)

  localWhizardLibFolder = os.path.join(localWhizardFolder, 'lib')
  if os.path.exists(localWhizardLibFolder):
    shutil.rmtree(localWhizardLibFolder)
  os.makedirs(localWhizardLibFolder)  # creates the lib folder

  whizardLibraries = getListOfLibraries(os.path.join(whizardResultFolder, "whizard"))
  copyLibsCall = ["rsync", "-avzL"]
  for lib in whizardLibraries:
    copyLibsCall.append(lib)
  copyLibsCall.append(localWhizardLibFolder)
  subprocess.Popen(copyLibsCall, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

  for fileName in folderlist:
    shutil.copy(fileName, localWhizardFolder)

  # Get the list of md5 sums for all the files in the folder to be tarred
  os.chdir(localWhizardFolder)
  subprocess.call(["find . -type f -exec md5sum {} > ../md5_checksum.md5 \\; && mv ../md5_checksum.md5 ."], shell=True)
  os.chdir(startDir)

  # Create the Tarball
  gLogger.notice("Creating Tarball...")
  appTar = localWhizardFolder + ".tgz"
  myappTar = tarfile.open(appTar, "w:gz")
  myappTar.add(localWhizardFolderRel)
  myappTar.close()

  md5sum = md5.md5(open(appTar, 'r').read()).hexdigest()

  gLogger.notice("...Done")

  gLogger.notice("Registering new Tarball in CS")
  tarballurl = {}

  av_platforms = gConfig.getSections(softwareSection, [])
  if av_platforms['OK']:
    if platform not in av_platforms['Value']:
      gLogger.error("Platform %s unknown, available are %s." % (platform, ", ".join(av_platforms['Value'])))
      gLogger.error("If yours is missing add it in CS")
      dexit(255)
  else:
    gLogger.error("Could not find all platforms available in CS")
    dexit(255)

  av_apps = gConfig.getSections("%s/%s" % (softwareSection, platform), [])
  if not av_apps['OK']:
    gLogger.error("Could not find all applications available in CS")
    dexit(255)

  if appName.lower() in av_apps['Value']:
    versions = gConfig.getSections("%s/%s/%s" % (softwareSection, platform, appName.lower()),
                                   [])
    if not versions['OK']:
      gLogger.error("Could not find all versions available in CS")
      dexit(255)
    if appVersion in versions['Value']:
      gLogger.error('Application %s %s for %s already in CS, nothing to do' % (appName.lower(), appVersion, platform))
      dexit(0)
    else:
      result = diracAdmin.csSetOption("%s/%s/%s/%s/TarBall" % (softwareSection, platform, appName.lower(), appVersion),
                                      os.path.basename(appTar))
      if result['OK']:
        modifiedCS = True
        tarballurl = gConfig.getOption("%s/%s/%s/TarBallURL" % (softwareSection, platform, appName.lower()), "")
        if len(tarballurl['Value']) > 0:
          res = upload(tarballurl['Value'], appTar)
          if not res['OK']:
            gLogger.error("Upload to %s failed" % tarballurl['Value'])
            dexit(255)
      result = diracAdmin.csSetOption("%s/%s/%s/%s/Md5Sum" % (softwareSection, platform, appName.lower(), appVersion),
                                      md5sum)
      if result['OK']:
        modifiedCS = True
      result = diracAdmin.csSetOption("%s/%s/%s/%s/Dependencies/beam_spectra/version" % (softwareSection,
                                                                                         platform,
                                                                                         appName.lower(),
                                                                                         appVersion),
                                      beam_spectra_version)

  else:
    result = diracAdmin.csSetOption("%s/%s/%s/%s/TarBall" % (softwareSection, platform,
                                                             appName.lower(), appVersion),
                                    os.path.basename(appTar))
    if result['OK']:
      modifiedCS = True
      tarballurl = gConfig.getOption("%s/%s/%s/TarBallURL" % (softwareSection, platform, appName.lower()),
                                     "")
      if len(tarballurl['Value']) > 0:
        res = upload(tarballurl['Value'], appTar)
        if not res['OK']:
          gLogger.error("Upload to %s failed" % tarballurl['Value'])
          dexit(255)

    result = diracAdmin.csSetOption("%s/%s/%s/%s/Md5Sum" % (softwareSection, platform, appName.lower(), appVersion),
                                    md5sum)

    result = diracAdmin.csSetOption("%s/%s/%s/%s/Dependencies/beam_spectra/version" % (softwareSection,
                                                                                       platform,
                                                                                       appName.lower(),
                                                                                       appVersion),
                                    beam_spectra_version)

  gLogger.verbose("Done uploading the tar ball")

  os.remove(appTar)

  # Set for all new processes the TarBallURL
  for process in inputlist.keys():
    inputlist[process]['TarBallCSPath'] = tarballurl['Value'] + os.path.basename(appTar)

  pl.updateProcessList(inputlist)

  pl.writeProcessList()

  input("Do you want to upload the process list? Press ENTER to proceed or CTRL-C to abort!")

  pl.uploadProcessListToFileCatalog(path_to_process_list, appVersion)

  # Commit the changes if nothing has failed and the CS has been modified
  if modifiedCS:
    result = diracAdmin.csCommitChanges(False)
    gLogger.verbose(result)
  gLogger.notice('All done OK!')
  dexit(0)


if __name__ == "__main__":
  main()
  dexit(0)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""This class is needed to get the Limited plugin support."""

from __future__ import absolute_import
__RCSID__ = "$Id$"
from DIRAC.TransformationSystem.Client.Transformation import Transformation as DT


class Transformation(DT):
  """ILCDirac version of Transformation."""

  def __init__(self, transID=0, transClient=None):
    super(Transformation, self).__init__(transID=0, transClient=None)
    self.supportedPlugins += ['Limited', 'Sliced', 'SlicedLimited']

  def _checkLimitedPlugin(self):
    """checks the Limitited Plugin, just passes the checkStandardPluging."""
    return self._checkStandardPlugin()

  def _checkSlicedPlugin(self):
    """checks the Sliced Plugin, just passes the checkStandardPluging."""
    return self._checkStandardPlugin()

  def _checkSlicedLimitedPlugin(self):
    """checks the SlicedLimited Plugin, just passes the checkStandardPluging."""
    return self._checkStandardPlugin()



#!/usr/local/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test Mokka module."""

from __future__ import print_function
from __future__ import absolute_import
import unittest
from mock import patch, MagicMock as Mock

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Interfaces.API.NewInterface.Applications import Mokka
from Tests.Utilities.GeneralUtils import assertEqualsImproved, assertDiracFailsWith, \
    assertDiracSucceeds

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Interfaces.API.NewInterface.Applications.Mokka'

# pylint: disable=protected-access


class MokkaTestCase(unittest.TestCase):
  """Base class for the Mokka test cases."""

  def setUp(self):
    """set up the objects."""
    self.mok = Mokka({})

  def test_setters(self):
    self.assertFalse(self.mok._errorDict)
    self.assertFalse(self.mok.inputSB)
    self.mok.setRandomSeed('invalid_seed')
    self.mok.setmcRunNumber(['something', False, []])
    self.mok.setDetectorModel({'bla': True})
    self.mok.setMacFile('lfn:/inval/dir/somefile.mac')
    self.mok.setStartFrom({'myset', False})
    self.mok.setProcessID(129843)
    self.mok.setDbSlice('lfn:/inval/dir/myDB.slice')
    print(self.mok._errorDict)
    assertEqualsImproved(len(self.mok._errorDict['_checkArgs']), 5, self)
    assertEqualsImproved(self.mok.inputSB, [
        'lfn:/inval/dir/somefile.mac', 'lfn:/inval/dir/myDB.slice'], self)
    assertEqualsImproved((self.mok.macFile, self.mok.dbSlice), (
        'lfn:/inval/dir/somefile.mac', 'lfn:/inval/dir/myDB.slice'), self)

  def test_setfiles_othercase(self):
    with patch('os.path.exists', new=Mock(return_value=False)):
      self.assertFalse(self.mok._errorDict)
      self.mok.setMacFile('/invalid/dir/myMac.file')
      self.mok.setDbSlice('/invalid/dir/someDb.sql')
      assertEqualsImproved((self.mok.macFile, self.mok.dbSlice),
                            ('/invalid/dir/myMac.file', '/invalid/dir/someDb.sql'), self)
      self.assertFalse(self.mok._errorDict)

  def test_userjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.mok._userjobmodules(module_mock), self)

  def test_prodjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.mok._prodjobmodules(module_mock), self)

  def test_userjobmodules_fails(self):
    with patch('%s._setUserJobFinalization' % MODULE_NAME, new=Mock(return_value=S_OK('something'))),\
        patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_test_err'))):
      assertDiracFailsWith(self.mok._userjobmodules(None),
                            'userjobmodules failed', self)

  def test_prodjobmodules_fails(self):
    with patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_OK('something'))), \
        patch('%s._setOutputComputeDataList' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_other_test_err'))):
      assertDiracFailsWith(self.mok._prodjobmodules(None),
                            'prodjobmodules failed', self)

  def test_checkconsistency(self):
    self.mok.version = 8431
    self.mok.steeringFile = 'lfn:/nonvalid/dir/mysteer.stdhep'
    self.mok._jobtype = 'User'
    assertDiracSucceeds(self.mok._checkConsistency(), self)

  def test_checkconsistency_nouserjob(self):
    with patch('os.path.exists', new=Mock(return_value=False)):
      self.mok.version = 8431
      self.mok.steeringFile = '/nonvalid/dir/mysteer.stdhep'
      self.mok._jobtype = 'notUser'
      self.mok.detectorModel = 'myTestDetv100'
      assertDiracSucceeds(self.mok._checkConsistency(), self)
      assertEqualsImproved(self.mok._listofoutput, [
          {"outputFile": "@{OutputFile}", "outputPath": "@{OutputPath}", "outputDataSE": '@{OutputSE}'}], self)
      assertEqualsImproved(self.mok.prodparameters, {
          'mokka_steeringfile': '/nonvalid/dir/mysteer.stdhep', 'mokka_detectormodel': 'myTestDetv100',
          'detectorType': 'ILD'}, self)

  def test_checkconsistency_nouserjob_2(self):
    with patch('os.path.exists', new=Mock(return_value=False)):
      self.mok.version = 8431
      self.mok.steeringFile = '/nonvalid/dir/mysteer.stdhep'
      self.mok._jobtype = 'notUser'
      self.mok.detectorModel = 0
      assertDiracSucceeds(self.mok._checkConsistency(), self)
      assertEqualsImproved(self.mok._listofoutput, [{
          "outputFile": "@{OutputFile}", "outputPath": "@{OutputPath}", "outputDataSE": '@{OutputSE}'
          }], self)
      assertEqualsImproved(self.mok.prodparameters, {
          'mokka_steeringfile': '/nonvalid/dir/mysteer.stdhep', 'detectorType': 'ILD'}, self)

  def test_checkconsistency_noversion(self):
    self.mok.version = None
    assertDiracFailsWith(self.mok._checkConsistency(), 'no version found', self)

  def test_checkconsistency_nosteeringfile(self):
    self.mok.version = True
    self.mok.steeringFile = None
    assertDiracFailsWith(self.mok._checkConsistency(), 'no steering file', self)

  def test_resolvelinkedparams(self):
    step_mock = Mock()
    input_mock = Mock()
    input_mock.getType.return_value = {'abc': False}
    self.mok._linkedidx = 3
    self.mok._jobsteps = [None, None, None, input_mock]
    assertDiracSucceeds(self.mok._resolveLinkedStepParameters(step_mock), self)
    step_mock.setLink.assert_called_once_with('InputFile', {'abc': False}, 'OutputFile')

  def test_resolvelinkedparams_noinputstep(self):
    self.mok._linkedidx = None
    self.mok._inputappstep = []
    assertDiracSucceeds(self.mok._resolveLinkedStepParameters(None), self)



#!/usr/bin/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the Whizard2 WorkflowModule."""

from __future__ import print_function
from __future__ import absolute_import
import unittest
import os
import os.path
import shutil
import tempfile
from mock import patch, MagicMock as Mock, mock_open

from parameterized import parameterized, param

from DIRAC import gLogger, S_OK, S_ERROR
from ILCDIRAC.Workflow.Modules.Whizard2Analysis import Whizard2Analysis
from Tests.Utilities.GeneralUtils import assertDiracSucceeds

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Workflow.Modules.Whizard2Analysis'
MODULEBASE_NAME = 'ILCDIRAC.Workflow.Modules.ModuleBase'
PROXYINFO_NAME = 'DIRAC.Core.Security.ProxyInfo'
# pylint: disable=too-many-public-methods, protected-access

gLogger.setLevel("ERROR")
gLogger.showHeaders(True)


def cleanup(tempdir):
  """Remove files after run."""
  try:
    shutil.rmtree(tempdir)
  except OSError:
    pass


@patch("%s.getProxyInfoAsString" % MODULEBASE_NAME, new=Mock(return_value=S_OK()))
@patch("%s.getProxyInfoAsString" % PROXYINFO_NAME, new=Mock(return_value=S_OK()))
class TestWhizard2Analysis(unittest.TestCase):
  """test Whizard2Analysis."""

  def assertIn(self, *args, **kwargs):
    """make this existing to placate pylint."""
    return super(TestWhizard2Analysis, self).assertIn(*args, **kwargs)

  @patch("%s.getProxyInfoAsString" % MODULEBASE_NAME, new=Mock(return_value=S_OK()))
  @patch("%s.getProxyInfoAsString" % PROXYINFO_NAME, new=Mock(return_value=S_OK()))
  def setUp(self):
    self.whiz = Whizard2Analysis()
    self.curdir = os.getcwd()
    self.tempdir = tempfile.mkdtemp("", dir="./")
    os.chdir(self.tempdir)
    self.whiz.ops = Mock()

  def tearDown(self):
    os.chdir(self.curdir)
    cleanup(self.tempdir)


class TestWhizard2AnalysisRunit(TestWhizard2Analysis):
  """test Whizard2 runtIt."""

  def setUp(self):
    super(TestWhizard2AnalysisRunit, self).setUp()
    self.logFileName = "python101.log"
    with open(self.logFileName, "w") as logF:
      logF.write("logged the logging logs")

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_Whizard2_runIt_success(self):
    """Whizard2.runit ................................................................................."""
    self.whiz.platform = 'Windows'
    self.whiz.applicationLog = self.logFileName
    # side effect for Script, userlibs, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[False, False, False, True])):
      res = self.whiz.runIt()
    print(res)
    assertDiracSucceeds(res, self)

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_Whizard2_runIt_failure_LogFile(self):
    """Whizard2.runit failure with applicationLog......................................................"""
    self.whiz.platform = "Windows"
    self.whiz.applicationLog = self.logFileName
    self.whiz.ignoreapperrors = False
    # side effect for Script, userlibs, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[False, False, False, False])):
      res = self.whiz.runIt()
    self.assertIn("did not produce the expected log", res['Message'])

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_Whizard2_runIt_failure_LogFile_ignore(self):
    """Whizard2.runit failure with applicationLog but ignore..........................................."""
    self.whiz.platform = "Windows"
    self.whiz.applicationLog = self.logFileName
    self.whiz.ignoreapperrors = True
    # side effect for Script, userlibs, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[False, False, False, False])):
      res = self.whiz.runIt()
    assertDiracSucceeds(res, self)

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_Whizard2_runIt_failure_NoLogFile(self):
    """Whizard2.runit failure with applicationLog not set............................................."""
    self.whiz.platform = "Windows"
    self.whiz.ignoreapperrors = True
    # side effect for Script, userlibs, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[False, False, False, False])):
      res = self.whiz.runIt()
    self.assertIn("No Log file provide", res['Message'])

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_Whizard2_runIt_failure_NoPlatform(self):
    """Whizard2.runit failure with platform ........................................................."""
    self.whiz.applicationLog = self.logFileName
    self.whiz.ignoreapperrors = True
    # side effect for Script, userlibs, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[False, False, False, False])):
      res = self.whiz.runIt()
    self.assertIn("No ILC platform selected", res['Message'])

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_Whizard2_runIt_success_LogAndScriptPresent(self):
    """Whizard2.runit success log and script exist..................................................."""
    self.whiz.platform = "Windows"
    self.whiz.applicationLog = self.logFileName
    self.whiz.ignoreapperrors = True
    with open("Whizard2__Run_.sh", "w") as scr:
      scr.write("content")
    with open("Whizard2__Steer_.sin", "w") as scr:
      scr.write("content")
    with open(self.logFileName, "w") as scr:
      scr.write("content")
    # side effect for Script, userlibs, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[True, True, False, True])):
      res = self.whiz.runIt()
    assertDiracSucceeds(res, self)

  @parameterized.expand([('NumberOfEvents', 100, 'n_events = 100'),
                         ('OutputFile', 'test.slcio', 'sample_format = lcio'),
                         ('OutputFile', 'test.ascii', 'sample_format = ascii'),
                         ('OutputFile', 'test.stdhep', 'sample_format = stdhep'),
                         ('randomSeed', '321', 'seed = 321'),
                         param('whizard2RawSin', True, 'sample_format', pUnExpected=True),
                         ])
  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_Whizard2_runIt_success_sinFile(self, pName, pValue, pExpected, pUnExpected=None):
    """Whizard.runit success with steeringFile........................................................."""
    self.whiz.platform = "Windows"
    self.whiz.applicationLog = self.logFileName
    self.whiz.whizard2SinFile = "whizard instructions"
    setattr(self.whiz, pName, pValue)
    # side effect for Steering1, Steering2, Script, userlib, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[False, False, False, True])):
      res = self.whiz.runIt()
    assertDiracSucceeds(res, self)
    self.assertEqual(self.whiz.whizard2SinFile, "whizard instructions")
    self.assertIn("whizard instructions", open("Whizard2__Steer_.sin").read())
    if not pUnExpected:
      self.assertIn(pExpected, open("Whizard2__Steer_.sin").read())
    else:
      self.assertNotIn(pExpected, open("Whizard2__Steer_.sin").read())

  @parameterized.expand([(S_ERROR(), S_OK(), 'Whizard2 should not proceed as'),
                         (S_OK(), S_ERROR('resInt Failed'), 'resInt Failed')])
  def test_Whizard2_runIt_fail(self, preStatus, resInt, errorMessage):
    """Whizard.runit fail steps......................................................................."""
    self.whiz.platform = "Windows"
    self.whiz.applicationLog = self.logFileName
    self.whiz.resolveIntegratedProcess = Mock(return_value=resInt)
    self.whiz.workflowStatus = preStatus
    res = self.whiz.runIt()
    if preStatus['OK']:
      self.assertIn(errorMessage, res['Message'])
    else:
      self.assertIn(errorMessage, res['Value'])

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_ERROR("missing setup.sh")))
  def test_Whizard2_runIt_fail_env(self):
    """Whizard.runit failed to get env................................................................"""
    self.whiz.platform = "Windows"
    self.whiz.applicationLog = self.logFileName
    res = self.whiz.runIt()
    self.assertEqual(res['Message'], "missing setup.sh")


class TestWhizard2AnalysisASI(TestWhizard2Analysis):
  """Whizard.ApplicationSpecificInputs."""

  @patch.dict(os.environ, {"JOBID": "12345"})
  def test_Whizard2_ASI_NoVariables(self):
    """Whizard.applicationSpecificInputs: checks that no variables have been set after this call......"""
    gLogger.setLevel("ERROR")
    self.whiz.workflow_commons = dict()
    self.whiz.applicationSpecificInputs()
    self.assertFalse(self.whiz.jobReport or self.whiz.productionID)

  @patch.dict(os.environ, {"JOBID": "12345"})
  def test_Whizard2_ASI_RandomSeed_Prod(self):
    """Whizard.applicationSpecificInputs: check setting of randomseed in production..................."""
    gLogger.setLevel("ERROR")
    self.whiz.workflow_commons = dict(IS_PROD=True, PRODUCTION_ID=6666, JOB_ID=123)
    self.whiz.OutputFile = 'events.stdhep'
    self.whiz.resolveInputVariables()
    self.whiz.applicationSpecificInputs()
    self.assertEqual(self.whiz.randomSeed, 6666123)

  @patch.dict(os.environ, {"JOBID": "12345"})
  def test_Whizard2_ASI_RandomSeed_Set(self):
    """Whizard.applicationSpecificInputs: check setting of default randomseed in user jobs............"""
    gLogger.setLevel("ERROR")
    self.whiz = Whizard2Analysis()
    self.whiz.workflow_commons = dict()
    self.whiz.resolveInputVariables()
    self.whiz.applicationSpecificInputs()
    self.assertEqual(int(self.whiz.randomSeed), 12345)

  @patch.dict(os.environ, {"JOBID": "12345"})
  def test_Whizard2_ASI_RandomSeed_User(self):
    """Whizard.applicationSpecificInputs: check setting of randomseed in user jobs...................."""
    gLogger.setLevel("ERROR")
    self.whiz = Whizard2Analysis()
    self.whiz.randomSeed = 654321
    self.whiz.workflow_commons = dict()
    self.whiz.resolveInputVariables()
    self.whiz.applicationSpecificInputs()
    self.assertEqual(int(self.whiz.randomSeed), 654321)

  @patch.dict(os.environ, {"JOBID": "12345"})
  def test_Whizard2_ASI_RandomSeed_User_Zero(self):
    """Whizard.applicationSpecificInputs: check setting of randomseed to zero in user jobs............"""
    gLogger.setLevel("ERROR")
    self.whiz = Whizard2Analysis()
    self.whiz.randomSeed = 0
    self.whiz.workflow_commons = dict()
    self.whiz.resolveInputVariables()
    self.whiz.applicationSpecificInputs()
    self.assertEqual(int(self.whiz.randomSeed), 0)

  def test_Whizard2_resolveIntegratedProcess_NoProc(self):
    """Test resolveIntegratedProcess with no integrated process."""
    gLogger.setLevel('ERROR')
    self.whiz = Whizard2Analysis()
    self.whiz.integratedProcess = ''
    ret = self.whiz.resolveIntegratedProcess()
    self.assertTrue(ret['OK'])

  GOD_RET_VALs = (S_OK({'tt': 'tt.tar'}), S_OK(dict(CVMFSPath='/c/c/c', TarBallURL='/ilc/vo')))

  @parameterized.expand([(False, (S_ERROR('No Processes'),), (None,), False),  # no processes defined
                         (False, (S_OK(), S_ERROR('No Options')), (None,), False),  # no options defined
                         (True, GOD_RET_VALs, (S_OK(),), True),  # cvmfs file exists
                         (True, GOD_RET_VALs, (S_OK(),), False),  # cvmfs file does not exist
                         (False, GOD_RET_VALs, (S_ERROR(),), False),  # getFile fails
                       ])
  def test_Whizard2_resolveIntegratedProcess_WithProc(self, success, sideEffects, gf_SE, localFile):
    self.whiz.integratedProcess = 'tt'
    self.whiz.ops = Mock()
    self.whiz.datMan = Mock(name='DatMan')
    self.whiz.datMan.getFile.side_effect = gf_SE
    self.whiz.ops.getOptionsDict.side_effect = sideEffects
    pathMock = Mock(name='pathMock')
    pathMock.path.join = os.path.join
    pathMock.path.exists.return_value = localFile
    with patch('%s.os' % MODULE_NAME, new=pathMock), \
         patch('%s.extractTarball' % MODULE_NAME, return_value=S_OK()):
      ret = self.whiz.resolveIntegratedProcess()
    self.assertEqual(ret['OK'], success)


def runTests():
  """Runs our tests."""
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestWhizard2Analysis)
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestWhizard2AnalysisRunit))
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestWhizard2AnalysisASI))
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)
  print(testResult)


if __name__ == '__main__':
  runTests()
